from __future__ import print_function

import itertools
import time
from datetime import datetime
from typing import (
    Dict,
    List,
)

from mxcubecore.HardwareObjects.abstract.ISPyBValueFactory import ISPyBValueFactory
from mxcubecore.utils.conversion import string_types

try:
    from urllib2 import URLError
    from urlparse import urljoin
except Exception:
    # Python3
    from urllib.error import URLError

import logging
import sys

from suds import WebFault
from suds.client import Client
from suds.sudsobject import asdict

from mxcubecore.model.lims_session import (
    LimsSessionManager,
    Proposal,
    Session,
)

suds_encode = str.encode

if sys.version_info > (3, 0):
    suds_encode = bytes.decode


def utf_encode(res_d):
    for key, value in res_d.items():
        if isinstance(value, dict):
            utf_encode(value)

        try:
            # Decode bytes object or encode str object depending
            # on Python version
            res_d[key] = suds_encode("utf8", "ignore")
        except Exception:
            # If not primitive or Text data, complext type, try to convert to
            # dict or str if the first fails
            try:
                res_d[key] = utf_encode(asdict(value))
            except Exception:
                try:
                    res_d[key] = str(value)
                except Exception:
                    res_d[key] = "ISPyBClient: could not encode value"

    return res_d


def utf_decode(res_d):
    for key, value in res_d.items():
        if isinstance(value, dict):
            utf_decode(value)
        try:
            res_d[key] = value.decode("utf8", "ignore")
        except Exception:
            pass

    return res_d


class ISPyBDataAdapter:
    def __init__(
        self,
        ws_root: str,
        proxy: dict,
        ws_username: str,
        ws_password: str,
        beamline_name: str,
    ):
        self.ws_root = ws_root
        self.ws_username = ws_username
        self.ws_password = ws_password
        self.proxy = proxy  # type: ignore
        self.beamline_name = beamline_name

        # the duration of the session in days for the ones that are created by MXCuBE
        self.new_sesssion_duration_days = 2

        self.logger = logging.getLogger("ispyb_adapter")

        self._shipping = self.__create_client(
            self.ws_root + "ToolsForShippingWebService?wsdl"
        )
        self._collection = self.__create_client(
            self.ws_root + "ToolsForCollectionWebService?wsdl"
        )
        self._tools_ws = self.__create_client(
            self.ws_root + "ToolsForBLSampleWebService?wsdl"
        )

    def __create_client(self, url: str):
        """
        Given a url it will create
        """
        if self.ws_root.strip().startswith("https://"):
            from suds.transport.https import HttpAuthenticated
        else:
            from suds.transport.http import HttpAuthenticated

        client = Client(
            url,
            timeout=3,
            transport=HttpAuthenticated(
                username=self.ws_username,  # type: ignore
                password=self.ws_password,
                proxy=self.proxy,
            ),
            cache=None,
            proxy=self.proxy,
        )
        client.set_options(cache=None, location=url)
        return client

    def isEnabled(self) -> object:
        return self._shipping  # type: ignore

    def create_session(self, proposal_id: str, beamline_name: str) -> Session:

        try:

            # proposal = self.find_proposal_by_login_and_beamline(self.get_user_, beamline_name)  # type: ignore
            current_time = time.localtime()
            start_time = time.strftime("%Y-%m-%d 00:00:00", current_time)
            end_time = (
                time.mktime(current_time)
                + 60 * 60 * 24 * self.new_sesssion_duration_days
            )
            tomorrow = time.localtime(end_time)
            end_time = time.strftime("%Y-%m-%d 07:59:59", tomorrow)

            session = {}
            session["proposalId"] = proposal_id
            session["beamlineName"] = beamline_name
            session["scheduled"] = 0
            session["nbShifts"] = 3
            session["comments"] = "Session created by the BCM"
            current_time = datetime.now()
            session["startDate"] = datetime.strptime(start_time, "%Y-%m-%d %H:%M:%S")
            session["endDate"] = datetime.strptime(end_time, "%Y-%m-%d %H:%M:%S")

            # return data to original codification
            # logging.getLogger("ispyb_client").info("Session creation: %s" % session)
            session_id = self._collection.service.storeOrUpdateSession(
                utf_decode(session)
            )
            logging.getLogger("ispyb_client").info(
                "Session created. session_id=%s" % session_id
            )
            response = self._collection.service.findSession(session_id)
            return self.__to_session(asdict(response))
        except Exception:
            raise

    def __is_scheduled_now(self, startDate: datetime, endDate: datetime) -> bool:
        return self.__is_time_between(startDate, endDate)

    def __is_time_between(
        self, start_date: datetime, end_date: datetime, check_time=None
    ):
        if start_date is None or end_date is None:
            return False

        begin_time = start_date.date()
        end_time = end_date.date()

        # If check time is not given, default to current UTC time
        check_time = check_time or datetime.utcnow().date()
        if begin_time <= check_time <= end_time:
            return True
        else:
            return False

    def __to_session(self, session: Dict[str, str]) -> Session:
        """
        Converts a dictionary composed by the person entries to the object proposal
        """
        proposal_name = session.get("proposalName")
        proposal_code = "".join(
            itertools.takewhile(lambda c: not c.isdigit(), proposal_name)
        )
        proposal_number = proposal_name[len(proposal_code) :]
        return Session(
            code=proposal_code,
            number=proposal_number,
            proposal_name=proposal_name,
            session_id=session.get("sessionId"),
            proposal_id=session.get("proposalId"),
            beamline_name=session.get("beamlineName"),
            comments=session.get("comments"),
            end_date=datetime.strftime(
                session.get("endDate"), "%Y%m%d"
            ),  # session.get("endDate"),
            nb_shifts=session.get("nbShifts"),
            scheduled=session.get("scheduled"),
            start_date=datetime.strftime(
                session.get("startDate"), "%Y%m%d"
            ),  # session.get("startDate"),
            start_datetime=session.get("startDate"),
            end_datetime=session.get("endDate"),
            is_scheduled_time=self.__is_scheduled_now(
                session.get("startDate"), session.get("endDate")
            ),
            is_scheduled_beamline=True,
        )

    def __to_proposal(self, proposal: Dict[str, str]) -> Proposal:
        """
        Converts a dictionary composed by the person entries to the object proposal
        """
        return Proposal(
            code=proposal.get("code").lower(),
            number=proposal.get("number").lower(),
            proposal_id=proposal.get("proposalId"),
            title=proposal.get("title"),
            type=proposal.get("type"),
        )

    def _debug(self, msg: str):
        logging.getLogger("HWR").debug(msg)

    def _exception(self, msg: str):
        logging.getLogger("HWR").exception(msg)

    def find_session(self, session_id: str) -> Session:
        try:
            response = self._collection.service.findSession(session_id)
            return self.__to_session(asdict(response))
        except Exception as e:
            self._exception(str(e))
            raise e

    def find_proposal(self, code: str, number: str) -> Proposal:
        try:
            self._debug("find_proposal. code=%s number=%s" % (code, number))
            response = self._shipping.service.findProposal(code, number)  # type: ignore
            return self.__to_proposal(asdict(response))  # type: ignore
        except Exception as e:
            self._exception(str(e))
            raise e

    def find_proposal_by_login_and_beamline(
        self, username: str, beamline_name: str
    ) -> Proposal:
        try:
            self._debug(
                "find_proposal_by_login_and_beamline. username=%s beamline_name=%s"
                % (username, beamline_name)
            )
            response = self._shipping.service.findProposalByLoginAndBeamline(
                username, beamline_name
            )
            print(response)
            if response is None:
                return []

            return self.__to_proposal(asdict(response))  # type: ignore
        except Exception as e:
            self._exception(str(e))
            raise e

    def find_sessions_by_proposal_and_beamLine(
        self, code: str, number: str, beamline: str
    ) -> List[Session]:
        try:
            self._debug(
                "find_sessions_by_proposal_and_beamLine. code=%s number=%s beamline=%s"
                % (code, number, beamline)
            )

            responses = self._collection.service.findSessionsByProposalAndBeamLine(
                code.upper(), number, beamline
            )
            sessions: List[Session] = []
            for response in responses:
                sessions.append(self.__to_session(asdict(response)))
            return sessions
        except Exception as e:
            self._exception(str(e))
            # raise e
        return []

    def _is_session_scheduled_today(self, session: Session) -> bool:
        now = datetime.now()
        if session.start_datetime.date() <= now.date() <= session.end_datetime.date():
            return True
        return False

    def get_sessions_by_code_and_number(
        self, code: str, number: str, beamline_name: str
    ) -> LimsSessionManager:
        try:
            self._debug(
                "get_sessions_by_code_and_number. code=%s number=%s beamline_name=%s"
                % (code, number, beamline_name)
            )
            sessions = self.find_sessions_by_proposal_and_beamLine(
                code, number, beamline_name
            )
            return LimsSessionManager(sessions=sessions)
        except WebFault as e:
            self._exception(str(e))
            raise e

    def get_person_by_username(self, username: str) -> Dict:
        try:
            person = self._shipping.service.findPersonByLogin(username)
            return asdict(person)
        except WebFault as e:
            self._error(str(e))

        return {}

    def get_sessions_by_username(
        self, username: str, beamline_name: str
    ) -> LimsSessionManager:
        try:
            self._debug(
                "get_sessions_by_username. username=%s beamline_name=%s"
                % (username, beamline_name)
            )

            proposal = self.find_proposal_by_login_and_beamline(username, beamline_name)  # type: ignore
            sessions = self.find_sessions_by_proposal_and_beamLine(
                proposal.code, proposal.number, beamline_name
            )
            return LimsSessionManager(
                sessions=sessions,
            )
        except WebFault as e:
            self._exception(str(e))
        return LimsSessionManager()

    ############# Legacy methods #####################
    def _store_data_collection_group(self, group_data):
        return self._collection.service.storeOrUpdateDataCollectionGroup(group_data)

    def store_data_collection_group(self, mx_collection):
        group_id = None
        # logging.getLogger("HWR").debug(
        #    "ispyb_group_data_collections: %s"
        #    % mx_collection["ispyb_group_data_collections"]
        # )

        if mx_collection["ispyb_group_data_collections"]:
            group_id = mx_collection.get("group_id", None)
        # logging.getLogger("HWR").debug(
        #    "Storing data collection group in lims. data to store. group_id: %s mx_collection: %s"
        #    % (str(group_id), str(mx_collection))
        # )
        # Create a new group id
        group = ISPyBValueFactory().dcg_from_dc_params(self._collection, mx_collection)
        # if group_id is None:
        group_id = self._collection.service.storeOrUpdateDataCollectionGroup(group)
        mx_collection["group_id"] = group_id

    def _update_data_collection(self, mx_collection):
        if "collection_id" in mx_collection:
            try:
                # Update the data collection group
                self.store_data_collection_group(mx_collection)
                data_collection = ISPyBValueFactory().from_data_collect_parameters(
                    self._collection, mx_collection
                )
                self._collection.service.storeOrUpdateDataCollection(data_collection)
            except WebFault as e:
                logging.getLogger("ispyb_client").exception(e)
            except URLError as e:
                logging.getLogger("ispyb_client").exception(e)
        else:
            logging.getLogger("ispyb_client").error(
                "Error in update_data_collection: "
                + "collection-id missing, the ISPyB data-collection is not updated."
            )

        return (0, 0)

    def store_image(self, image_dict):
        if self._collection:
            logging.getLogger("HWR").debug("Storing image in lims")
            if "dataCollectionId" in image_dict:
                try:
                    image_id = self._collection.service.storeOrUpdateImage(image_dict)
                    logging.getLogger("HWR").debug(
                        "  - storing image in lims ok. id : %s" % image_id
                    )
                    return image_id
                except WebFault:
                    logging.getLogger("ispyb_client").exception(
                        "ISPyBClient: exception in store_image"
                    )
                except URLError as e:
                    logging.getLogger("ispyb_client").exception(e)
            else:
                logging.getLogger("ispyb_client").error(
                    "Error in store_image: "
                    + "data_collection_id missing, could not store image in ISPyB"
                )
        else:
            logging.getLogger("ispyb_client").exception(
                "Error in store_image: could not connect to server"
            )

    def get_samples(self, proposal_id):
        response_samples = []

        if self._tools_ws:
            try:
                response_samples = (
                    self._tools_ws.service.findSampleInfoLightForProposal(
                        proposal_id, self.beamline_name
                    )
                )
                response_samples = [
                    utf_encode(asdict(sample)) for sample in response_samples
                ]

            except WebFault as e:
                logging.getLogger("ispyb_client").exception(str(e))
            except URLError as e:
                logging.getLogger("ispyb_client").exception(e)
        else:
            logging.getLogger("ispyb_client").exception(
                "Error in get_samples: could not connect to server"
            )

        return response_samples

    def store_robot_action(self, robot_action_dict):
        """Stores robot action"""
        logging.getLogger("HWR").debug("Storing robot actions in lims")

        if True:
            robot_action_vo = self._collection.factory.create("robotActionWS3VO")

            robot_action_vo.actionType = robot_action_dict.get("actionType")
            robot_action_vo.containerLocation = robot_action_dict.get(
                "containerLocation"
            )
            robot_action_vo.dewarLocation = robot_action_dict.get("dewarLocation")

            # robot_action_vo.endTime = robot_action_dict.get("endTime")
            robot_action_vo.message = robot_action_dict.get("message")
            robot_action_vo.sampleBarcode = robot_action_dict.get("sampleBarcode")
            robot_action_vo.sessionId = robot_action_dict.get("sessionId")
            robot_action_vo.blSampleId = robot_action_dict.get("sampleId")
            robot_action_vo.startTime = datetime.strptime(
                robot_action_dict.get("startTime"), "%Y-%m-%d %H:%M:%S"
            )
            robot_action_vo.endTime = datetime.strptime(
                robot_action_dict.get("endTime"), "%Y-%m-%d %H:%M:%S"
            )
            robot_action_vo.status = robot_action_dict.get("status")
            robot_action_vo.xtalSnapshotAfter = robot_action_dict.get(
                "xtalSnapshotAfter"
            )
            robot_action_vo.xtalSnapshotBefore = robot_action_dict.get(
                "xtalSnapshotBefore"
            )
            return self._collection.service.storeRobotAction(robot_action_vo)
        return None

    def associate_bl_sample_and_energy_scan(self, entry_dict):
        try:
            return self._collection.service.storeBLSampleHasEnergyScan(
                entry_dict["energyScanId"], entry_dict["blSampleId"]
            )
        except Exception as e:
            logging.getLogger("ispyb_client").exception(str(e))
            return -1

    def get_data_collection(self, data_collection_id):
        try:
            dc_response = self._collection.service.findDataCollection(
                data_collection_id
            )

            dc = utf_encode(asdict(dc_response))
            dc["startTime"] = datetime.strftime(dc["startTime"], "%Y-%m-%d %H:%M:%S")
            dc["endTime"] = datetime.strftime(dc["endTime"], "%Y-%m-%d %H:%M:%S")
            return dc
        except Exception as e:
            logging.getLogger("ispyb_client").exception(str(e))
            return {}

    def find_detector(self, type, manufacturer, model, mode):
        """
        Returns the Detector3VO object with the characteristics
        matching the ones given.
        """
        if self._collection:
            try:
                res = self._collection.service.findDetectorByParam(
                    "", manufacturer, model, mode
                )
                return res
            except WebFault:
                logging.getLogger("ispyb_client").exception(
                    "ISPyBClient: exception in find_detector"
                )
        else:
            logging.getLogger("ispyb_client").exception(
                "Error find_detector: could not connect to" + " server"
            )

    def update_session(self, session_dict):
        if self._collection:

            try:
                print(session_dict)
                # The old API used date formated strings and the new
                # one uses DateTime objects.
                session_dict["startDate"] = datetime.strptime(
                    session_dict["startDate"], "%Y-%m-%d %H:%M:%S"
                )
                session_dict["endDate"] = datetime.strptime(
                    session_dict["endDate"], "%Y-%m-%d %H:%M:%S"
                )

                try:
                    session_dict["lastUpdate"] = datetime.strptime(
                        session_dict["lastUpdate"].split("+")[0], "%Y-%m-%d %H:%M:%S"
                    )
                    session_dict["timeStamp"] = datetime.strptime(
                        session_dict["timeStamp"].split("+")[0], "%Y-%m-%d %H:%M:%S"
                    )
                except Exception:
                    pass

                # return data to original codification
                decoded_dict = utf_decode(session_dict)
                session = self._collection.service.storeOrUpdateSession(decoded_dict)

                # changing back to string representation of the dates,
                # since the session_dict is used after this method is called,
                session_dict["startDate"] = datetime.strftime(
                    session_dict["startDate"], "%Y-%m-%d %H:%M:%S"
                )
                session_dict["endDate"] = datetime.strftime(
                    session_dict["endDate"], "%Y-%m-%d %H:%M:%S"
                )

            except WebFault as e:
                session = {}
                logging.getLogger("ispyb_client").exception(str(e))
            except URLError:
                logging.getLogger("ispyb_client").exception(_CONNECTION_ERROR_MSG)

            # logging.getLogger("ispyb_client").info(
            #    "[ISPYB] Session goona be created: session_dict %s" % session_dict
            # )
            logging.getLogger("ispyb_client").info(
                "[ISPYB] Session created: %s" % session
            )
            return session
        else:
            logging.getLogger("ispyb_client").exception(
                "Error in create_session: could not connect to server"
            )

    def store_beamline_setup(self, session_id, bl_config):
        blSetupId = None
        if self._collection:

            session = {}

            try:
                session = self.get_session(session_id)
            except Exception:
                logging.getLogger("ispyb_client").exception(
                    "ISPyBClient: exception in store_beam_line_setup"
                )
            else:
                if session is not None:
                    try:
                        blSetupId = self._collection.service.storeOrUpdateBeamLineSetup(
                            bl_config
                        )

                        session["beamLineSetupId"] = blSetupId
                        self.update_session(session)

                    except WebFault as e:
                        logging.getLogger("ispyb_client").exception(str(e))
                    except URLError:
                        logging.getLogger("ispyb_client").exception(
                            _CONNECTION_ERROR_MSG
                        )
        else:
            logging.getLogger("ispyb_client").exception(
                "Error in store_beamline_setup: could not connect" + " to server"
            )

        return blSetupId

    def store_data_collection(self, mx_collection, bl_config=None):
        logging.getLogger("HWR").info("Storing datacollection in ISPyB")
        return self._store_data_collection(mx_collection, bl_config=bl_config)

    def update_data_collection(self, mx_collection):
        logging.getLogger("HWR").info("Updating datacollection in ISPyB")
        return self._update_data_collection(mx_collection)

    def finalize_data_collection(self, mx_collection):
        logging.getLogger("HWR").info("Updating datacollection in ISPyB")
        return self._update_data_collection(mx_collection)

    def _store_data_collection(self, mx_collection, bl_config=None):
        data_collection = ISPyBValueFactory().from_data_collect_parameters(
            self._collection, mx_collection
        )

        detector_id = 0
        if bl_config:
            lims_beamline_setup = ISPyBValueFactory.from_bl_config(
                self._collection, bl_config
            )

            lims_beamline_setup.synchrotronMode = data_collection.synchrotronMode

            self.store_beamline_setup(mx_collection["sessionId"], lims_beamline_setup)

            detector_params = ISPyBValueFactory().detector_from_blc(
                bl_config, mx_collection
            )

            detector = self.find_detector(*detector_params)
            detector_id = 0

            if detector:
                detector_id = detector.detectorId
                data_collection.detectorId = detector_id

        collection_id = self._collection.service.storeOrUpdateDataCollection(
            data_collection
        )
        logging.getLogger("HWR").debug(
            "Storing data collection ok. collection id : %s" % collection_id
        )

        return (collection_id, detector_id)

    def get_session(self, session_id):
        try:
            logging.getLogger("HWR").debug("get_session. session_id=%s" % session_id)
            session = self._collection.service.findSession(session_id)
            logging.getLogger("HWR").debug("get_session. session=%s" % session)
            if session is not None:
                session.startDate = datetime.strftime(
                    session.startDate, "%Y-%m-%d %H:%M:%S"
                )
                session.endDate = datetime.strftime(
                    session.endDate, "%Y-%m-%d %H:%M:%S"
                )
                return utf_encode(asdict(session))
        except Exception as e:
            logging.getLogger("ispyb_client").exception(e)

        return {}

    def store_energy_scan(self, energyscan_dict):
        status = {"energyScanId": -1}

        try:
            energyscan_dict["startTime"] = datetime.strptime(
                energyscan_dict["startTime"], "%Y-%m-%d %H:%M:%S"
            )

            energyscan_dict["endTime"] = datetime.strptime(
                energyscan_dict["endTime"], "%Y-%m-%d %H:%M:%S"
            )

            try:
                del energyscan_dict["remoteEnergy"]
            except KeyError:
                pass

            status["energyScanId"] = self._collection.service.storeOrUpdateEnergyScan(
                energyscan_dict
            )

        except Exception as e:
            logging.getLogger("ispyb_client").exception(str(e))

        return status

    def store_xfe_spectrum(self, xfespectrum_dict):
        status = {"xfeFluorescenceSpectrumId": -1}
        try:
            if isinstance(xfespectrum_dict["startTime"], string_types):
                xfespectrum_dict["startTime"] = datetime.strptime(
                    xfespectrum_dict["startTime"], "%Y-%m-%d %H:%M:%S"
                )

                xfespectrum_dict["endTime"] = datetime.strptime(
                    xfespectrum_dict["endTime"], "%Y-%m-%d %H:%M:%S"
                )
            else:
                xfespectrum_dict["startTime"] = xfespectrum_dict["startTime"]
                xfespectrum_dict["endTime"] = xfespectrum_dict["endTime"]

            status["xfeFluorescenceSpectrumId"] = (
                self._collection.service.storeOrUpdateXFEFluorescenceSpectrum(
                    xfespectrum_dict
                )
            )

        except URLError as e:
            logging.getLogger("ispyb_client").exception(str(e))

        return status

    def update_bl_sample(self, bl_sample):
        if self._disabled:
            return {}

        if self._tools_ws:
            try:
                status = self._tools_ws.service.storeOrUpdateBLSample(bl_sample)
            except WebFault as e:
                logging.getLogger("ispyb_client").exception(str(e))
                status = {}
            except URLError:
                logging.getLogger("ispyb_client").exception(
                    "Could not connect to ISPyB, URLerror"
                )

            return status
        else:
            logging.getLogger("ispyb_client").exception(
                "Error in update_bl_sample: could not connect to server"
            )



"""
Session hardware object.

Contains information regarding the current session and methods to
access and manipulate this information.
"""

import os
import socket
import time
from typing import Tuple

from mxcubecore.BaseHardwareObjects import HardwareObject
from mxcubecore.model.queue_model_objects import PathTemplate

default_raw_data_folder = "RAW_DATA"
default_processed_data_folder = "PROCESSED_DATA"
default_archive_folder = "ARCHIVE"


class Session(HardwareObject):
    def __init__(self, name):
        HardwareObject.__init__(self, name)

        self.synchrotron_name = None
        self.beamline_name = None

        self.session_id = None
        self.proposal_code = None
        self.proposal_number = None
        self.proposal_id = None
        self.in_house_users = []
        self.endstation_name = None
        self.session_start_date = None
        self.user_group = ""
        self.email_extension = None
        self.template = None

        self.default_precision = 5
        self.suffix = None

        self.base_directory = None
        self.base_process_directory = None
        self.base_archive_directory = None

        self.raw_data_folder_name = default_raw_data_folder
        self.processed_data_folder_name = default_processed_data_folder

    # Framework-2 method, inherited from HardwareObject and called
    # by the framework after the object has been initialized.
    def init(self):
        self.synchrotron_name = self.get_property("synchrotron_name")
        self.beamline_name = self.get_property("beamline_name")
        self.endstation_name = self.get_property("endstation_name").lower()

        self.suffix = self["file_info"].get_property("file_suffix")
        self.template = self["file_info"].get_property("file_template")

        base_directory = self["file_info"].get_property("base_directory")

        base_process_directory = self["file_info"].get_property(
            "processed_data_base_directory"
        )

        base_archive_directory = self["file_info"].get_property(
            "archive_base_directory"
        )

        folder_name = self["file_info"].get_property("raw_data_folder_name")
        if folder_name and folder_name.strip():
            self.raw_data_folder_name = folder_name

        folder_name = self["file_info"].get_property("processed_data_folder_name")
        if folder_name and folder_name.strip():
            self.processed_data_folder_name = folder_name

        archive_folder = self["file_info"].get_property("archive_folder")
        if archive_folder:
            archive_folder = archive_folder.strip()
        if not archive_folder:
            archive_folder = default_archive_folder
        try:
            inhouse_proposals = self["inhouse_users"]["proposal"]
            for prop in inhouse_proposals:
                self.in_house_users.append(
                    (prop.get_property("code"), str(prop.get_property("number")))
                )
        except KeyError:
            pass

        email_extension = self.get_property("email_extension")
        if email_extension:
            self.email_extension = email_extension
        else:
            try:
                domain = socket.getfqdn().split(".")
                self.email_extension = ".".join((domain[-2], domain[-1]))
            except (TypeError, IndexError):
                pass

        self.set_base_data_directories(
            base_directory,
            base_process_directory,
            base_archive_directory,
            raw_folder=self.raw_data_folder_name,
            process_folder=self.processed_data_folder_name,
            archive_folder=archive_folder,
        )

        try:
            precision = int(self["file_info"].get_property("precision", ""))
        except ValueError:
            precision = self.default_precision

        PathTemplate.set_precision(precision)
        PathTemplate.set_path_template_style(self.synchrotron_name, self.template)

    def set_base_data_directories(
        self,
        base_directory,
        base_process_directory,
        base_archive_directory,
        raw_folder="RAW_DATA",
        process_folder="PROCESSED_DATA",
        archive_folder="ARCHIVE",
    ):
        self.base_directory = base_directory
        self.base_process_directory = base_process_directory
        self.base_archive_directory = base_archive_directory

        self.raw_data_folder_name = raw_folder
        self.processed_data_folder_name = process_folder

        if self.base_directory is not None:
            PathTemplate.set_data_base_path(self.base_directory)

        if self.base_archive_directory is not None:
            PathTemplate.set_archive_path(self.base_archive_directory, archive_folder)

    def get_base_data_directory(self):
        """
        Returns the base data directory taking the 'contextual'
        information into account, such as if the current user
        is inhouse.

        :returns: The base data path.
        :rtype: str
        """
        user_category = ""
        directory = ""

        if self.session_start_date:
            # start_time = self.session_start_date.split(" ")[0].replace("-", "")
            start_time = (
                self.session_start_date
            )  # .strftime("%Y%m%d")  # .split(" ")[0].replace("-", "")
        else:
            start_time = time.strftime("%Y%m%d")

        if self.is_inhouse():
            user_category = "inhouse"
            directory = os.path.join(
                self.base_directory,
                self.endstation_name,
                user_category,
                self.get_proposal(),
                start_time,
            )
        else:
            user_category = "visitor"
            directory = os.path.join(
                self.base_directory,
                user_category,
                self.get_proposal(),
                self.endstation_name,
                start_time,
            )

        return directory

    def get_path_with_proposal_as_root(self, path: str) -> str:
        """
        Strips the beginning of the path so that it starts with
        the proposal folder as root

        :path: The full path
        :returns: Path stripped so that it starts with proposal
        """
        if self.is_inhouse():
            user_category = "inhouse"
            directory = os.path.join(
                self.base_directory, self.endstation_name, user_category
            )
        else:
            user_category = "visitor"
            directory = os.path.join(self.base_directory, user_category)

        return path.split(directory)[1]

    def get_base_image_directory(self):
        """
        :returns: The base path for images.
        :rtype: str
        """
        return os.path.join(self.get_base_data_directory(), self.raw_data_folder_name)

    def get_base_process_directory(self):
        """
        :returns: The base path for processed data.
        :rtype: str
        """
        return os.path.join(
            self.get_base_data_directory(), self.processed_data_folder_name
        )

    def get_image_directory(self, sub_dir: str = "") -> str:
        """
        Returns the full path to images

        :param subdir: sub directory relative to path returned
                       by get_base_image_directory

        :returns: The full path to images.
        """
        directory = self.get_base_image_directory()

        if sub_dir:
            sub_dir = sub_dir.replace(" ", "").replace(":", "-")
            directory = os.path.join(directory, sub_dir) + os.path.sep

        return directory

    def get_process_directory(self, sub_dir: str = "") -> str:
        """
        Returns the full path to processed data,

        :param subdir: sub directory relative to path returned
                       by get_base_process_directory

        :returns: The full path to processed data.
        """
        directory = self.get_base_process_directory()

        if sub_dir:
            sub_dir = sub_dir.replace(" ", "").replace(":", "-")
            directory = os.path.join(directory, sub_dir) + "/"

        return directory

    def get_full_path(self, subdir: str = "", tag: str = "") -> Tuple[str, str]:
        """
        Returns the full path to both image and processed data.
        The path(s) returned will follow the convention:

          <base_directory>/<subdir>/run_<NUMBER>_<tag>

        Where NUMBER is a automatically sequential number and
        base_directory the path returned by get_base_image/process_directory

        :param subdir: subdirectory
        :param tag: tag for

        :returns: Tuple with the full path to image and processed data
        """

        return self.get_image_directory(subdir), self.get_process_directory(subdir)

    def get_default_prefix(self, sample_data_node=None, generic_name=False):
        """
        Returns the default prefix, using sample data such as the
        acronym as parts in the prefix.

        :param sample_data_node: The data node to get additional
                                 information from, (which will be
                                 added to the prefix).
        :type sample_data_node: Sample


        :returns: The default prefix.
        :rtype: str
        """
        proposal = self.get_proposal()
        prefix = proposal

        if sample_data_node:
            name = sample_data_node.name
            protein_acronym = sample_data_node.crystals[0].protein_acronym
            if protein_acronym:
                if name:
                    prefix = "%s-%s" % (protein_acronym, name)
                else:
                    prefix = protein_acronym
            else:
                prefix = name or ""
        elif generic_name:
            prefix = "<acronym>-<name>"

        return prefix

    def get_default_subdir(self, sample_data: dict) -> str:
        """
        Gets the default sub-directory based on sample information

        Args:
           sample_data: Lims sample dictionary

        Returns:
           Sub-directory path string
        """

        subdir = ""

        if isinstance(sample_data, dict):
            sample_name = sample_data.get("sampleName", "")
            protein_acronym = sample_data.get("proteinAcronym", "")
        else:
            sample_name = sample_data.name
            protein_acronym = sample_data.crystals[0].protein_acronym

        if protein_acronym:
            subdir = "%s/%s-%s/" % (protein_acronym, protein_acronym, sample_name)
        else:
            subdir = "%s/" % sample_name

        return subdir.replace(":", "-")

    def get_archive_directory(self):
        archive_directory = os.path.join(
            self["file_info"].get_property("archive_base_directory"),
            self["file_info"].get_property("archive_folder"),
        )

        archive_directory = PathTemplate.get_archive_directory()

        return archive_directory

    def get_proposal(self):
        """
        :returns: The proposal, 'local-user' if no proposal is
                  available
        :rtype: str
        """
        proposal = "local-user"

        if self.proposal_code and self.proposal_number:
            if self.proposal_code == "ifx":
                self.proposal_code = "fx"

            proposal = "%s%s" % (self.proposal_code, self.proposal_number)

        return proposal.lower().replace("-", "")

    def is_inhouse(self, proposal_code=None, proposal_number=None):
        """
        Determines if a given proposal is considered to be inhouse.

        :param proposal_code: Proposal code
        :type proposal_code: str

        :param proposal_number: Proposal number
        :type proposal_number: str

        :returns: True if the proposal is inhouse, otherwise False.
        :rtype: bool
        """
        if not proposal_code:
            proposal_code = self.proposal_code

        if not proposal_number:
            proposal_number = self.proposal_number

        if (proposal_code, proposal_number) in self.in_house_users:
            return True
        else:
            return False

    def get_inhouse_user(self):
        """
        :returns: The current inhouse user.
        :rtype: tuple (<proposal_code>, <proposal_number>)
        """
        return self.in_house_users[0]

    def set_session_start_date(self, start_date_str):
        """
        :param start_date_str: The session start date
        :type start_date_str: str
        """
        self.session_start_date = start_date_str

    def get_session_start_date(self):
        """
        :returns: The session start date
        :rtype: str
        """
        return self.session_start_date

    def set_user_group(self, group_name):
        """
        :param group_name: Name of user group
        :type group_name: str
        """
        self.user_group = str(group_name)

    def get_group_name(self):
        """
        :returns: Name of user group
        :rtype: str
        """
        return self.user_group



"""ESRF SC3 Sample Changer Hardware Object"""

import ESRFSC3
import SC3

from mxcubecore.TaskUtils import task


class Command:
    def __init__(self, cmd):
        self.cmd = cmd

    def isSpecConnected(self):
        return True

    @task
    def __call__(self, *args, **kwargs):
        self.cmd(*args, **kwargs)


class ID30SC3(ESRFSC3.ESRFSC3):
    def __init__(self, *args, **kwargs):
        ESRFSC3.ESRFSC3.__init__(self, *args, **kwargs)

    def init(self):
        controller = self.get_object_by_role("controller")

        SC3.SC3.init(self)
        self.prepareCentringAfterLoad = True

        self.prepareCentring = Command(controller.prepare_centring)
        self._moveToLoadingPosition = Command(
            controller.move_to_sample_loading_position
        )
        self._moveToUnloadingPosition = Command(
            controller.move_to_sample_loading_position
        )

    @task
    def unlockMinidiffMotors(self):
        pass



# encoding: utf-8
#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU General Lesser Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""Photon flux calculations
Example xml_ configuration:

.. code-block:: xml

 <object class="ESRF.ESRFPhotonFlux">
   <username>Photon flux</username>
   <object role="controller" href="/bliss"/>
   <object role="aperture" href="/udiff_aperture"/>
   <counter_name>i0</counter_name>
   <beam_check_name>checkbeam</beam_check_name>
   <object role="monitor_beam" href="/monitor_beam"/>
 </object>
"""
import logging

import gevent

from mxcubecore import HardwareRepository as HWR
from mxcubecore.HardwareObjects.abstract.AbstractFlux import AbstractFlux


class ESRFPhotonFlux(AbstractFlux):
    """Photon flux calculation for ID30B"""

    def __init__(self, name):
        super().__init__(name)
        self._counter = None
        self._flux_calc = None
        self._aperture = None
        self.threshold = None
        self._beam_check_obj = None
        self._monitorbeam_obj = None

    def init(self):
        """Initialisation"""
        super().init()
        controller = self.get_object_by_role("controller")

        self._aperture = self.get_object_by_role("aperture")
        self.threshold = self.get_property("threshold") or 0.0

        try:
            self._flux_calc = controller.CalculateFlux()
            self._flux_calc.init()
        except AttributeError:
            logging.getLogger("HWR").exception(
                "Could not get flux calculation from BLISS"
            )

        counter = self.get_property("counter_name")
        if counter:
            self._counter = getattr(controller, counter)
        else:
            self._counter = self.get_object_by_role("counter")

        beam_check = self.get_property("beam_check_name")

        if beam_check:
            self._beam_check_obj = getattr(controller, beam_check)

        self._monitorbeam_obj = self.get_object_by_role("monitor_beam")

        try:
            HWR.beamline.safety_shutter.connect("stateChanged", self.update_value)
        except AttributeError as err:
            raise RuntimeError("Safety shutter is not configured") from err

        self._poll_task = gevent.spawn(self._poll_flux)

    def _poll_flux(self):
        """Poll the flux every 2 seconds"""
        while True:
            self.re_emit_values()
            gevent.sleep(3)

    def get_value(self):
        """Calculate the flux value as function of a reading"""

        counts = self._counter.raw_read
        if isinstance(counts, list):
            counts = float(counts[0])
        counts = float(self._counter.raw_read)
        if counts == -9999:
            counts = 0.0

        try:
            egy = HWR.beamline.energy.get_value()
            calib = self._flux_calc.calc_flux_factor(egy * 1000.0)[
                self._counter.diode.name
            ]
        except AttributeError:
            egy = 0
            calib = 0

        try:
            label = self._aperture.get_value().name
            aperture_factor = self._aperture.get_factor(label)
            if isinstance(aperture_factor, tuple):
                factor = aperture_factor[0] + aperture_factor[1] * egy
            else:
                factor = float(aperture_factor)
        except (AttributeError, ValueError, RuntimeError):
            factor = 1.0

        counts = abs(counts * calib * factor)
        if counts < self.threshold:
            return 0.0

        return counts

    @property
    def check_beam(self):
        """Check if there is beam
        Returns:
            (bool): True if beam present, False otherwise
        """
        return self._beam_check_obj.check_beam()

    def wait_for_beam(self, timeout=None):
        """Wait until beam present
        Args:
            timeout (float): optional - timeout [s],
                             If timeout == 0: return at once and do not wait
                                              (default);
                             if timeout is None: wait forever.
        """
        if self._monitorbeam_obj:
            try:
                check = self._monitorbeam_obj.get_value().value
            except AttributeError:
                check = False
            if check is True:
                try:
                    return self._beam_check_obj.wait_for_beam(timeout)
                except AttributeError:
                    return True
        return True



#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

import logging
import os
import tempfile
import time
from datetime import datetime

import gevent

from mxcubecore import HardwareRepository as HWR
from mxcubecore.BaseHardwareObjects import HardwareObjectState
from mxcubecore.HardwareObjects.abstract import AbstractSampleChanger
from mxcubecore.HardwareObjects.abstract.sample_changer import (
    Container,
    Crims,
    Sample,
)

POSITION_DESC = {
    "Park": "Parked",
    "PickDew": "Pick from dewar",
    "PutDew": "Put in the dewar",
    "DewToMD": "On the way from dewar to MD",
    "MD": "MD",
    "Dryer": "Dryer",
    "CTB": "Center to base",
    "BTC": "Base to center",
    "CTEject": "Center puck eject",
    "EjectTC": "Put to cener puck",
}

STATUS_DESC = {
    "idl": "Idle",
    "bsy": "Busy",
    "err": "Error",
    "opn": "Opened",
    "on": "On",
    "off": "Off",
}

STATUS_STR_DESC = {
    "Sys": "Controller",
    "Rob": "Robot",
    "Grp": "Gripper",
    "Lid": "Dewar Lid",
    "Mag": "MD smart magnet",
    "Cry": "Cryo stream position",
    "Gui": "Guillotine",
    "Trsf": "Sample transfer state",
    "MD": "MD transfer state",
    "CDor": "Robot cage door",
    "CPuck": "Central puck",
    "VDet": "Vial in gripper detected",
    "Dry": "Dry gripper routine",
    "Prgs": "Progress bar",
    "PSw": "Puck switches",
    "SDet": "Sample detected on MD",
    "RPos": "Robot positions",
    "SPNr": "Sample puck in operation",
    "Err": "Error",
    "LErr": "Last 5 errors",
    "LSPmnt": "Last sample mounted",
    "CMD": "Command in progress",
    "MntPos": "MD mounting position",
    "Vial": "Vial detected",
}

CMD_STR_DESC = {
    "IDL": "Idle",
    "Nxt": "Nxt ?",
    "Mnt": "Mount sample",
    "Dis": "Dismount sample",
    "Tst": "Test ?",
    "Dry": "Dry",
}

ERROR_STR_DESC = {
    0: "No Error",
    1: "Guillotine valve 1",
    2: "Guillotine valve 2",
    3: "Puck switches",
    4: "Gripper",
    5: "Air pressure",
    6: "Lid valve 1",
    7: "Lid valve 2",
    8: "Crash",
    9: "Magnet",
    10: "Transfer",
    11: "Communication with diffractometer",
}


class Marvin(AbstractSampleChanger.SampleChanger):

    __TYPE__ = "Marvin"

    def __init__(self, *args, **kwargs):
        super(Marvin, self).__init__(self.__TYPE__, False, *args, **kwargs)
        self._selected_sample = None
        self._selected_basket = None

        self._num_baskets = None
        self._status_list = []
        self._state_string = None
        self._puck_switches = None
        self._centre_puck = None
        self._mounted_puck = None
        self._mounted_sample = None
        self._action_started = None
        self._progress = None
        self._veto = None
        self._sample_detected = None
        self._focusing_mode = None
        self._process_step_info = None
        self._command_list = None
        self._info_dict = {}
        self._in_error_state = False
        self._was_mount_error = False
        self._command_acknowledgement = False

        self.chan_status = None
        self.chan_sample_is_loaded = None
        self.chan_puck_switched = None
        self.chan_mounted_sample_puck = None
        self.chan_process_step_info = None

        self.cmd_mount_sample = None
        self.cmd_unmount_sample = None
        self.cmd_open_lid = None
        self.cmd_close_lid = None
        self.cmd_base_to_center = None
        self.cmd_center_to_base = None
        self.cmd_dry_gripper = None

        self.beam_focusing_hwobj = None

    def init(self):
        self._puck_switches = 0
        self._num_basket = self.get_property("numBaskets")
        if not self._num_basket:
            self._num_basket = 17

        for i in range(self._num_basket):
            basket = Container.Basket(self, i + 1)
            self._add_component(basket)

        self.chan_mounted_sample_puck = self.get_channel_object("chanMountedSamplePuck")
        self.chan_mounted_sample_puck.connect_signal(
            "update", self.mounted_sample_puck_changed
        )

        self.chan_process_step_info = self.get_channel_object(
            "chanProcessStepInfo", optional=True
        )
        if self.chan_process_step_info is not None:
            self.chan_process_step_info.connect_signal(
                "update", self.process_step_info_changed
            )

        self.chan_command_list = self.get_channel_object(
            "chanCommandList", optional=True
        )
        if self.chan_command_list is not None:
            self.chan_command_list.connect_signal("update", self.command_list_changed)

        self.chan_puck_switches = self.get_channel_object("chanPuckSwitches")
        self.chan_puck_switches.connect_signal("update", self.puck_switches_changed)

        self.chan_status = self.get_channel_object("chanStatusList")
        self.chan_status.connect_signal("update", self.status_list_changed)

        self.chan_sample_is_loaded = self.get_channel_object("chanSampleIsLoaded")
        self.chan_sample_is_loaded.connect_signal(
            "update", self.sample_is_loaded_changed
        )

        self.chan_veto = self.get_channel_object("chanVeto", optional=True)
        if self.chan_veto is not None:
            self.chan_veto.connect_signal("update", self.veto_changed)

        self.cmd_mount_sample = self.get_command_object("cmdMountSample")
        self.cmd_unmount_sample = self.get_command_object("cmdUnmountSample")
        self.cmd_open_lid = self.get_command_object("cmdOpenLid")
        self.cmd_close_lid = self.get_command_object("cmdCloseLid")
        self.cmd_base_to_center = self.get_command_object("cmdBaseToCenter")
        self.cmd_center_to_base = self.get_command_object("cmdCenterToBase")
        self.cmd_dry_gripper = self.get_command_object("cmdDryGripper")

        self.beam_focusing_hwobj = self.get_object_by_role("beam_focusing")
        if self.beam_focusing_hwobj is not None:
            self.connect(
                self.beam_focusing_hwobj,
                "focusingModeChanged",
                self.focusing_mode_changed,
            )
            (
                self._focusing_mode,
                beam_size,
            ) = self.beam_focusing_hwobj.get_active_focus_mode()
            self.focusing_mode_changed(self._focusing_mode, beam_size)
        else:
            self._focusing_mode = "P13mode"

        self._init_sc_contents()
        self._update_state()
        self._updateSCContents()
        self._update_loaded_sample()

        self.log_filename = self.get_property("log_filename")
        if self.log_filename is None:
            self.log_filename = os.path.join(
                tempfile.gettempdir(), "mxcube", "marvin.log"
            )
        logging.getLogger("HWR").debug("Marvin log filename: %s" % self.log_filename)
        AbstractSampleChanger.SampleChanger.init(self)

        self.update_state(HardwareObjectState.READY)
        self.status_list_changed(self.chan_status.get_value())
        self.puck_switches_changed(self.chan_puck_switches.get_value())
        self.mounted_sample_puck_changed(self.chan_mounted_sample_puck.get_value())
        self.sample_is_loaded_changed(self.chan_sample_is_loaded.get_value())

    def get_status_str_desc(self):
        return STATUS_STR_DESC

    def get_log_filename(self):
        """Returns log filename"""
        return self.log_filename

    def run_test(self):
        """Test method mounts/dismounts samples"""
        samples_mounted = 0
        for cycle in range(5):
            for sample_index in range(1, 11):
                logging.getLogger("GUI").info(
                    "Sample changer: Mounting sample 1:%d" % sample_index
                )
                self.load("1:%02d" % sample_index, wait=True)
                logging.getLogger("GUI").info(
                    "Sample changer: Total mounts done: %d" % (samples_mounted + 1)
                )
                samples_mounted += 1
                gevent.sleep(1)

    def puck_switches_changed(self, puck_switches):
        """Updates puck switches"""
        self._puck_switches = int(puck_switches)
        self._info_dict["puck_switches"] = int(puck_switches)
        self._updateSCContents()

    def sample_is_loaded_changed(self, sample_detected):
        """Updates sample is loaded"""
        if self._sample_detected != sample_detected:

            if sample_detected:
                logging.getLogger("HWR").debug("Sample changer: sample re-appeared")
            else:
                logging.getLogger("HWR").debug("Sample changer: sample disappeared")

            self._sample_detected = sample_detected
            self._info_dict["sample_detected"] = sample_detected
            self._update_loaded_sample()
            self.update_info()

    def wait_command_acknowledgement(self, timeout):
        with gevent.Timeout(
            timeout, Exception("Timeout waiting for command acknowldegement")
        ):
            logging.getLogger("HWR").debug(
                "Sample changer: start waiting command acknowldegement"
            )
            while not self._command_acknowledgement:
                gevent.sleep(0.05)
            logging.getLogger("HWR").debug(
                "Sample changer: done waiting command acknowldegement"
            )

    def wait_sample_to_disappear(self, timeout):
        with gevent.Timeout(
            timeout, Exception("Timeout waiting for sample to disappear")
        ):
            logging.getLogger("HWR").debug(
                "Sample changer: start waiting sample to disappear"
            )
            while self._sample_detected:
                if self._was_mount_error:
                    self._was_mount_error = False
                    return
                gevent.sleep(0.05)
            logging.getLogger("HWR").debug(
                "Sample changer: done  waiting sample to disappear"
            )

    def wait_sample_to_appear(self, timeout):
        with gevent.Timeout(timeout, Exception("Timeout waiting for sample to appear")):
            logging.getLogger("HWR").debug(
                "Sample changer: start waiting sample to appear"
            )
            while not self._sample_detected:
                if self._was_mount_error:
                    self._was_mount_error = False
                    return
                gevent.sleep(0.05)
            logging.getLogger("HWR").debug(
                "Sample changer: done  waiting sample to appear"
            )

    def wait_sample_on_gonio(self, timeout):
        # with gevent.Timeout(timeout, Exception("Timeout waiting for sample on gonio")):
        #    while not self._sample_detected:
        #        gevent.sleep(0.05)
        with gevent.Timeout(timeout, Exception("Timeout waiting for centring phase")):
            while (
                HWR.beamline.diffractometer.get_current_phase()
                != HWR.beamline.diffractometer.PHASE_CENTRING
            ):
                if not self._is_device_busy():
                    return
                gevent.sleep(0.05)

    def is_sample_on_gonio(self):
        return self.chan_sample_is_loaded.get_value()
        # logging.getLogger("GUI").info("Sample on gonio check 1: %s" %first_try)
        # gevent.sleep(1.0)
        # second_try = self.chan_sample_is_loaded.get_value()
        # logging.getLogger("GUI").info("Sample on gonio check 2: %s" %second_try)
        # return first_try and second_try

    def mounted_sample_puck_changed(self, mounted_sample_puck):
        """Updates mounted puck index"""
        mounted_sample = mounted_sample_puck[0] - 1
        mounted_puck = mounted_sample_puck[1] - 1

        if mounted_puck != self._mounted_puck:
            self._mounted_puck = mounted_puck
            if self._focusing_mode == "P13mode":
                self._info_dict["mounted_puck"] = mounted_puck
            else:
                self._info_dict["mounted_puck"] = mounted_puck + 1
            self._updateSCContents()

        if mounted_sample != self._mounted_sample:
            self._mounted_sample = mounted_sample
            if self._focusing_mode == "P13mode":
                self._info_dict["mounted_sample"] = mounted_sample
            else:
                self._info_dict["mounted_sample"] = mounted_sample + 1
            self._update_loaded_sample()

    def veto_changed(self, status):
        """Veto changed callback. Used to wait for ready"""
        self._veto = status
        self._info_dict["veto"] = self._veto

    def focusing_mode_changed(self, focusing_mode, beam_size):
        """Sets CRL combination based on the focusing mode"""
        self._focusing_mode = focusing_mode
        self._info_dict["focus_mode"] = self._focusing_mode

    def process_step_info_changed(self, process_step_info):
        self._process_step_info = process_step_info.replace("\n", " ")
        self._command_acknowledgement = True
        if "error" in process_step_info.lower():
            self._was_mount_error = True
            logging.getLogger("GUI").error(
                "Sample changer: %s" % self._process_step_info
            )
            # GB: 20190304: this seemd to lock mxcube forever on any marvin error
            # self._in_error_state = True
            # self._set_state(AbstractSampleChanger.SampleChangerState.Alarm)

        else:
            logging.getLogger("GUI").info(
                "Sample changer: %s" % self._process_step_info
            )
        self._info_dict["process_step"] = self._process_step_info

    def command_list_changed(self, cmd_list):
        self._command_list = cmd_list.replace("\n", "")
        logging.getLogger("GUI").info(
            "Sample changer: Last command - %s" % self._command_list
        )
        self._info_dict["command_list"] = self._command_list

    def open_lid(self):
        self.cmd_open_lid(1)

    def close_lid(self):
        self.cmd_close_lid(1)

    def base_to_center(self):
        return
        # self.cmd_base_to_center(1)

    def center_to_base(self):
        return
        # self.cmd_center_to_base(1)

    def dry_gripper(self):
        self.cmd_dry_gripper(1)

    def get_sample_properties(self):
        """Gets sample properties"""
        return (Container.Pin.__HOLDER_LENGTH_PROPERTY__,)

    def assert_can_execute_task(self):
        return

    def _do_update_info(self):
        """Updates the sample changers status: mounted pucks, state,
        currently loaded sample
        """
        # self._update_state()
        # self._updateSCContents()
        # call this method if status string changed
        # self._update_loaded_sample()

    def _directly_update_selected_component(self, basket_no, sample_no):
        """Directly updates necessary sample"""
        basket = None
        sample = None
        if basket_no is not None and basket_no > 0 and basket_no <= self._num_basket:
            basket = self.get_component_by_address(
                Container.Basket.get_basket_address(basket_no)
            )
            if (
                sample_no is not None
                and sample_no > 0
                and sample_no <= len(basket.get_sample_list())
            ):
                sample = self.get_component_by_address(
                    Container.Pin.get_sample_address(basket_no, sample_no)
                )
        self._set_selected_component(basket)
        self._set_selected_sample(sample)

    def _do_select(self, component):
        """Selects a new component (basket or sample).
        Uses method >_directly_update_selected_component< to actually
        search and select the corrected positions.
        """
        if type(component) in (Container.Pin, Sample.Sample):
            selected_basket_no = component.get_basket_no()
            selected_sample_no = component.get_index() + 1
        elif isinstance(component, Container.Container) and (
            component.get_type() == Container.Basket.__TYPE__
        ):
            selected_basket_no = component.get_index() + 1
            selected_sample_no = None

        self._directly_update_selected_component(selected_basket_no, selected_sample_no)

    def _do_scan(self, component, recursive):
        """Scans the barcode of a single sample, puck or recursively even the
        complete sample changer.
        Not implemented
        """
        print("_do_scan TODO")

    def _do_load(self, sample=None):
        """Loads a sample on the diffractometer. Performs a simple put operation
        if the diffractometer is empty, and a sample exchange (unmount of
        old + mount of  new sample) if a sample is already mounted on
        the diffractometer.
        """
        # self._set_state(AbstractSampleChanger.SampleChangerState.Ready)
        log = logging.getLogger("GUI")

        if self._focusing_mode not in (
            "Collimated",
            "Double",
            "Imaging",
            "TREXX",
            "P13mode",
        ):
            error_msg = "Focusing mode is undefined. Sample loading is disabled"
            log.error(error_msg)
            return

        # if self._focusing_mode in ("Collimated", "Double") and not self._centre_puck:
        #    log.error("No center puck detected. Please do Base-to-Center with any puck.")
        #    return

        if self._in_error_state:
            log.error(
                "Sample changer is in error state. "
                + "All commands are disabled."
                + "Fix the issue and reset sample changer in MXCuBE"
            )
            return

        start_time = datetime.now()
        selected = self.get_selected_sample()

        if sample is not None:
            if sample != selected:
                self._do_select(sample)
                selected = self.get_selected_sample()
        else:
            if selected is not None:
                sample = selected
            else:
                raise Exception("No sample selected")

        basket_index = selected.get_basket_no()
        sample_index = selected.get_vial_no()

        # 1. Check if sample is on gonio. This should never happen
        # because if mount is requested and on gonio is sample then
        # first sample is dismounted
        if self._focusing_mode == "P13mode":
            if self.is_sample_on_gonio():
                if selected == self.get_loaded_sample():
                    msg = (
                        "The sample "
                        + str(self.get_loaded_sample().get_address())
                        + " is already loaded"
                    )
                    raise Exception(msg)
                else:
                    self._do_unload()

        msg = "Sample changer: Loading sample %d:%d" % (
            int(basket_index),
            int(sample_index),
        )
        log.warning(msg + " Please wait...")
        self.emit("progressInit", (msg, 100, False))

        # 2. Set diffractometer transfer phase
        logging.getLogger("HWR").debug(
            "%s %s"
            % (
                HWR.beamline.diffractometer.get_current_phase(),
                HWR.beamline.diffractometer.PHASE_TRANSFER,
            )
        )
        if (
            HWR.beamline.diffractometer.get_current_phase()
            != HWR.beamline.diffractometer.PHASE_TRANSFER
        ):
            logging.getLogger("HWR").debug("set transfer")
            HWR.beamline.diffractometer.set_phase(
                HWR.beamline.diffractometer.PHASE_TRANSFER, 60.0
            )
            time.sleep(2)
            if (
                HWR.beamline.diffractometer.get_current_phase()
                != HWR.beamline.diffractometer.PHASE_TRANSFER
            ):
                log.error(
                    "Diffractometer is not in the transfer phase. "
                    + "Sample will not be mounted"
                )
                raise Exception("Unable to set Transfer phase")

        # logging.getLogger("HWR").debug("Sample changer: Closing guillotine...")
        # HWR.beamline.detector.close_cover()
        # logging.getLogger("HWR").debug("Sample changer: Guillotine closed")
        # 3. If necessary move detector to save position
        if self._focusing_mode == "P13mode":
            if HWR.beamline.detector.distance.get_value() < 399.0:
                log.info("Sample changer: Moving detector to save position...")
                self._veto = 1
                HWR.beamline.detector.distance.set_value(400, timeout=45)
                time.sleep(1)
                self.waitVeto(20.0)
                log.info("Sample changer: Detector moved to save position")
        else:
            pass
            # logging.getLogger("HWR").debug("Sample changer: Closing guillotine...")
            # HWR.beamline.detector.close_cover()
            ##logging.getLogger("HWR").debug("Sample changer: Guillotine closed")

        # 4. Executed command and wait till device is ready
        if self._focusing_mode == "P13mode":
            self._execute_server_task(
                self.cmd_mount_sample, int(sample_index), int(basket_index)
            )
        else:
            if (
                self._focusing_mode == "Collimated"
                or self._focusing_mode == "Imaging"
                or self._focusing_mode == "TREXX"
            ):
                self._execute_server_task(
                    self.cmd_mount_sample, int(sample_index), int(basket_index), 1
                )
            elif self._focusing_mode == "Double":
                self._execute_server_task(
                    self.cmd_mount_sample, int(sample_index), int(basket_index), 3
                )

        self.emit("progressStop", ())

        if self.is_sample_on_gonio():
            log.info(
                "Sample changer: Sample %d:%d loaded"
                % (int(basket_index), int(sample_index))
            )
            if self._focusing_mode == "P13mode":
                HWR.beamline.diffractometer.set_phase(
                    HWR.beamline.diffractometer.PHASE_CENTRING, 60.0
                )
                # HWR.beamline.diffractometer.close_kappa()
        else:
            log.error(
                "Sample changer: Failed to load sample %d:%d"
                % (int(basket_index), int(sample_index))
            )
            raise Exception("Sample not loaded!")

    def load(self, sample=None, wait=True):
        """Load a sample"""
        # self._set_state(AbstractSampleChanger.SampleChangerState.Ready)
        if self._focusing_mode == "P13mode":
            AbstractSampleChanger.SampleChanger.load(self, sample, wait)
        else:
            sample = self._resolve_component(sample)
            self.assert_not_charging()
            return self._execute_task(
                AbstractSampleChanger.SampleChangerState.Loading,
                wait,
                self._do_load,
                sample,
            )

    def _do_unload(self, sample_slot=None):
        """Unloads a sample from the diffractometer"""
        log = logging.getLogger("GUI")

        # self._set_state(AbstractSampleChanger.SampleChangerState.Ready)
        if self._focusing_mode not in (
            "Collimated",
            "Double",
            "Imaging",
            "TREXX",
            "P13mode",
        ):
            error_msg = "Focusing mode is undefined. Sample loading is disabled"
            log.error(error_msg)
            return

        if self._in_error_state:
            log.error(
                "Sample changer is in error state. "
                + "All commands are disabled."
                + "Fix the issue and reset sample changer in MXCuBE"
            )
            return

        if self._focusing_mode == "P13mode":
            sample_index = self._mounted_sample
            basket_index = self._mounted_puck
        else:
            sample_index = self._mounted_sample + 1
            basket_index = self._mounted_puck + 1

        msg = "Sample changer: Unloading sample %d:%d" % (basket_index, sample_index)
        log.warning(msg + ". Please wait...")
        self.emit("progressInit", (msg, 100, False))

        if (
            HWR.beamline.diffractometer.get_current_phase()
            != HWR.beamline.diffractometer.PHASE_TRANSFER
        ):
            HWR.beamline.diffractometer.set_phase(
                HWR.beamline.diffractometer.PHASE_TRANSFER, 60
            )
            if (
                HWR.beamline.diffractometer.get_current_phase()
                != HWR.beamline.diffractometer.PHASE_TRANSFER
            ):
                log.error(
                    "Diffractometer is not in the transfer phase. "
                    + "Sample will not be mounted"
                )
                raise Exception("Unable to set Transfer phase")

        # HWR.beamline.detector.close_cover()
        if self._focusing_mode == "P13mode":
            if HWR.beamline.detector.distance.get_value() < 399.0:
                log.info("Sample changer: Moving detector to save position ...")
                self._veto = 1
                HWR.beamline.detector.distance.set_value(400, timeout=45)
                time.sleep(1)
                self.waitVeto(20.0)
                log.info("Sample changer: Detector moved to save position")
        else:
            pass
            # HWR.beamline.detector.close_cover()

        start_time = datetime.now()

        if self._focusing_mode == "P13mode":
            self._execute_server_task(
                self.cmd_unmount_sample, sample_index, basket_index
            )
        else:
            if (
                self._focusing_mode == "Collimated"
                or self._focusing_mode == "Imaging"
                or self._focusing_mode == "TREXX"
            ):

                self._execute_server_task(
                    self.cmd_unmount_sample, sample_index, basket_index, 1
                )
            elif self._focusing_mode == "Double":
                self._execute_server_task(
                    self.cmd_unmount_sample, sample_index, basket_index, 3
                )

        self.emit("progressStop", ())

        if self.is_sample_on_gonio():
            log.error(
                "Sample changer: Failed to unload sample %d:%d"
                % (basket_index, sample_index)
            )
            raise Exception("Sample not unloaded!")
        else:
            log.info(
                "Sample changer: Sample %d:%d unloaded" % (basket_index, sample_index)
            )

    def clear_basket_info(self, basket):
        """Clears information about basket"""
        # TODO
        return

    def _do_change_mode(self, mode):
        """Changes the mode of sample changer"""
        return

    def _do_abort(self):
        """Aborts the sample changer"""
        return

    def _do_reset(self):
        """Clean all sample info, move sample to his position and move puck
        from center to base"""
        self._set_state(AbstractSampleChanger.SampleChangerState.Ready)
        self._init_sc_contents()
        self._in_error_state = False

    def _execute_server_task(self, method, *args):
        """Executes called cmd, waits until sample changer is ready and
        updates loaded sample info
        """
        # self.wait_ready(60.0)
        self._state_string = "Bsy"
        self._progress = 5

        arg_arr = []
        for arg in args:
            arg_arr.append(arg)

        logging.getLogger("HWR").debug(
            "Sample changer: Sending cmd with arguments: %s..." % str(arg_arr)
        )

        self._command_acknowledgement = False

        method(arg_arr)
        logging.getLogger("HWR").debug("Sample changer: Waiting ready...")
        self.wait_command_acknowledgement(5.0)
        self._action_started = True
        gevent.sleep(5)
        if method == self.cmd_mount_sample:
            # self.wait_sample_on_gonio(120.0)
            self._was_mount_error = False
            self.wait_sample_to_disappear(40.0)
            self.wait_sample_to_appear(60.0)
        else:
            self.wait_ready(120.0)
        logging.getLogger("HWR").debug("Sample changer: Ready")
        logging.getLogger("HWR").debug("Sample changer: Waiting veto...")
        self.waitVeto(20.0)
        logging.getLogger("HWR").debug("Sample changer: Veto ready")
        # if self._is_device_busy():
        #    raise Exception("Action finished to early. Sample changer is not ready!!!")
        self.sample_is_loaded_changed(self.chan_sample_is_loaded.get_value())
        # self._update_state()
        self._update_loaded_sample()
        # self._set_state(AbstractSampleChanger.SampleChangerState.Ready)
        self.update_state(HardwareObjectState.READY)
        self._action_started = False

    def _update_state(self):
        state = self._read_state()
        if state == HardwareObjectState.BUSY and self._is_device_busy(self.get_state()):
            return
        self._set_state(state)

    def _read_state(self):
        """Converts state string to defined state"""
        state_converter = {
            "ALARM": HardwareObjectState.FAULT,
            "Err": HardwareObjectState.FAULT,
            "Idl": HardwareObjectState.READY,
            "Bsy": HardwareObjectState.BUSY,
        }
        return state_converter.get(self._state_string, HardwareObjectState.UNKNOWN)

    def _is_device_busy(self, state=None):
        """Checks whether Sample changer is busy"""
        if state is None:
            state = self._read_state()
        if self._progress >= 100 and state in (
            HardwareObjectState.READY,
            HardwareObjectState.FAULT,
        ):
            return False
        else:
            return True

    def _is_device_ready(self):
        """Checks whether Sample changer is ready"""
        state = self._read_state()
        return state in (HardwareObjectState.READY,)

    def wait_ready(self, timeout=None):
        """Waits until the samle changer is ready"""
        with gevent.Timeout(timeout, Exception("Timeout waiting for device ready")):
            while self._is_device_busy():
                gevent.sleep(0.05)

    def waitVeto(self, timeout=20):
        with gevent.Timeout(timeout, Exception("Timeout waiting for veto")):
            while self._veto == 1:
                self.veto_changed(self.chan_veto.get_value)
                gevent.sleep(0.1)

    def _update_selection(self):
        """Updates selected basked and sample"""

        basket = None
        sample = None
        try:
            basket_no = self._selected_basket
            if (
                basket_no is not None
                and basket_no > 0
                and basket_no <= self._num_basket
            ):
                basket = self.get_component_by_address(
                    Container.Basket.get_basket_address(basket_no)
                )
                sample_no = self._selected_sample
                if (
                    sample_no is not None
                    and sample_no > 0
                    and sample_no <= Container.Basket.NO_OF_SAMPLES_PER_PUCK
                ):
                    sample = self.get_component_by_address(
                        Container.Pin.get_sample_address(basket_no, sample_no)
                    )
        except Exception:
            pass
        self._set_selected_component(basket)
        self._set_selected_sample(sample)

    def _update_loaded_sample(self):
        """
        Updates loaded sample
        """

        if (
            self._sample_detected
            and self._mounted_sample > -1
            and self._mounted_puck > -1
            and self._centre_puck
        ):
            if self._focusing_mode == "P13mode":
                new_sample = self.get_component_by_address(
                    Container.Pin.get_sample_address(
                        self._mounted_puck, self._mounted_sample
                    )
                )
            else:
                new_sample = self.get_component_by_address(
                    Container.Pin.get_sample_address(
                        self._mounted_puck + 1, self._mounted_sample + 1
                    )
                )
        else:
            new_sample = None

        if self.get_loaded_sample() != new_sample:
            old_sample = self.get_loaded_sample()
            if old_sample is not None:
                # there was a sample on the gonio
                loaded = False
                has_been_loaded = True
                old_sample._set_loaded(loaded, has_been_loaded)
            if new_sample is not None:
                self._update_sample_barcode(new_sample)
                loaded = True
                has_been_loaded = True
                new_sample._set_loaded(loaded, has_been_loaded)

    def _update_sample_barcode(self, sample):
        """
        Updates the barcode of >sample< in the local database
        after scanning with the barcode reader.
        """
        datamatrix = "NotAvailable"
        scanned = len(datamatrix) != 0
        if not scanned:
            datamatrix = "----------"
        sample._set_info(sample.is_present(), datamatrix, scanned)

    def _init_sc_contents(self):
        """
        Initializes the sample changer content with default values.
        """
        basket_list = [("", 4)] * self._num_basket
        for basket_index in range(self._num_basket):
            basket = self.get_components()[basket_index]
            datamatrix = None
            present = scanned = True
            basket._set_info(present, datamatrix, scanned)

        # create temporary list with default sample information and indices
        sample_list = []
        for basket_index in range(self._num_basket):
            for sample_index in range(10):
                sample_list.append(
                    (
                        "",
                        basket_index + 1,
                        sample_index + 1,
                        1,
                        Container.Pin.STD_HOLDERLENGTH,
                    )
                )
        # write the default sample information into permanent Pin objects
        for spl in sample_list:
            sample = self.get_component_by_address(
                Container.Pin.get_sample_address(spl[1], spl[2])
            )
            datamatrix = None
            present = scanned = loaded = has_been_loaded = False
            sample._set_info(present, datamatrix, scanned)
            sample._set_loaded(loaded, has_been_loaded)
            sample._set_holder_length(spl[4])

    def _updateSCContents(self):
        """
        Updates sample changer content
        """
        for basket_index in range(self._num_basket):
            basket = self.get_components()[basket_index]

            if self._focusing_mode == "P13mode":
                bsk_index = basket_index + 1
            else:
                bsk_index = basket_index

            if (
                (int(self._puck_switches) & pow(2, basket_index) > 0)
                or (self._mounted_puck == bsk_index)
                and self._centre_puck
            ):
                # f puck_switches & (1 << basket_index):
                # basket was mounted
                present = True
                scanned = False
                datamatrix = None
            else:
                # basket was removed
                present = False
                scanned = False
                datamatrix = None

            basket._set_info(present, datamatrix, scanned)
            # set the information for all dependent samples
            """
            for sample_index in range(10):
                sample = self.get_component_by_address(Pin.get_sample_address(\
                    (basket_index + 1), (sample_index + 1)))
                present = sample.get_container().is_present()
                if present:
                    datamatrix = '%d:%d - Not defined' % \
                       (bsk_index, sample_index)
                else:
                    datamatrix = None
                datamatrix = None
                scanned = False
                sample._set_info(present, datamatrix, scanned)
                # forget about any loaded state in newly mounted or removed basket)
                loaded = _has_been_loaded = False
                sample._set_loaded(loaded, has_been_loaded)
            """

        self._trigger_selection_changed_event()

    def status_list_changed(self, status_string):
        tmp_string = status_string.replace(" ", "")
        tmp_string1 = tmp_string.replace("On\r", "On")
        status_string = tmp_string1.replace("\rSys", ";Sys")
        self._status_list = status_string.split(";")

        for status in self._status_list:
            property_status_list = status.split(":")
            if len(property_status_list) < 2:
                continue

            prop_name = property_status_list[0]
            prop_value = property_status_list[1]

            if prop_name == "Rob":
                if (
                    self._state_string != prop_value
                    and prop_value in ("Idl", "Bsy", "Err")
                    and self._action_started
                ):
                    self._state_string = prop_value
                    logging.getLogger("HWR").debug(
                        "Sample changer: status changed: %s" % self._state_string
                    )
                    self._update_state()
            elif prop_name == "Prgs":
                try:
                    if int(prop_value) != self._progress and self._action_started:
                        self._progress = int(prop_value)
                        self.emit("progressStep", self._progress)
                        self._info_dict["progress"] = self._progress
                except Exception:
                    pass
            elif prop_name == "CPuck":
                if prop_value == "1":
                    centre_puck = True
                elif prop_value == "0":
                    centre_puck = False

                if centre_puck != self._centre_puck:
                    self._centre_puck = centre_puck
                    self._info_dict["centre_puck"] = self._centre_puck
                    self._updateSCContents()
                    self._update_loaded_sample()
            elif prop_name == "Lid":
                self._info_dict["lid_opened"] = prop_value == "Opn"
            elif prop_name == "Err":
                logging.getLogger("GUI").error(
                    "Sample changer: Error (%s)" % prop_value
                )
                logging.getLogger("GUI").error("Details: ")

                for status in self._status_list:
                    property_status_list = status.split(":")
                    if len(property_status_list) < 2:
                        continue

                    prop_name = property_status_list[0]
                    prop_value = property_status_list[1]

                    if prop_name in STATUS_STR_DESC:
                        logging.getLogger("GUI").error(
                            " - %s: %s " % (STATUS_STR_DESC[prop_name], prop_value)
                        )

        self.emit("statusListChanged", self._status_list)
        self.emit("infoDictChanged", self._info_dict)

    def force_emit_signals(self):
        self.emit("statusListChanged", self._status_list)
        self.emit("infoDictChanged", self._info_dict)
        self._trigger_info_changed_event()
        self._trigger_selection_changed_event()



# encoding: utf-8
#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""TangoShutter class - interface for shutter controlled by TANGO
Implements _set_value, get_value methods

Tango states are:
"UNKNOWN", "OPEN", "CLOSED", "FAULT", "MOVING", "DISABLE", "STANDBY", "RUNNING"

Example xml file:
<object class = "TangoShutter">
  <username>Safety Shutter</username>
  <tangoname>ab/cd/ef</tangoname>
  <command type="tango" name="Open">Open</command>
  <command type="tango" name="Close">Close</command>
  <channel type="tango" name="State" polling="1000">State</channel>
  <values>{"open": "OPEN", "cloded": "CLOSED", "DISABLE" : "DISABLE"}</values>
</object>

In this example the <values> tag contains a json dictionary that maps spectific tango shutter states to the
convantional states defined in the TangoShutter Class. This tag is not necessay in cases where the tango shutter states
are all covered by the TangoShuter class conventional states.
"""
import json
from enum import (
    Enum,
    unique,
)

from mxcubecore.BaseHardwareObjects import HardwareObjectState
from mxcubecore.HardwareObjects.abstract.AbstractShutter import AbstractShutter

__copyright__ = """ Copyright © 2023 by the MXCuBE collaboration """
__license__ = "LGPLv3+"


@unique
class TangoShutterStates(Enum):
    """Shutter states definitions."""

    CLOSED = HardwareObjectState.READY, "CLOSED"
    OPEN = HardwareObjectState.READY, "OPEN"
    MOVING = HardwareObjectState.BUSY, "MOVING"
    DISABLE = HardwareObjectState.WARNING, "DISABLE"
    AUTOMATIC = HardwareObjectState.READY, "RUNNING"
    UNKNOWN = HardwareObjectState.UNKNOWN, "RUNNING"
    FAULT = HardwareObjectState.WARNING, "FAULT"


class TangoShutter(AbstractShutter):
    """TANGO implementation of AbstractShutter"""

    SPECIFIC_STATES = TangoShutterStates

    def __init__(self, name):
        super().__init__(name)
        self.open_cmd = None
        self.close_cmd = None
        self.state_channel = None

    def init(self):
        """Initilise the predefined values"""
        super().init()
        self.open_cmd = self.get_command_object("Open")
        self.close_cmd = self.get_command_object("Close")
        self.state_channel = self.get_channel_object("State")
        self._initialise_values()
        self.state_channel.connect_signal("update", self._update_value)
        self.update_state()

        try:
            self.config_values = json.loads(self.get_property("values"))
        except:
            self.config_values = None

    def _update_value(self, value):
        """Update the value.
        Args:
            value(str): The value reported by the state channel.
        """
        if self.config_values:
            value = self.config_values[str(value)]
        else:
            value = str(value)

        super().update_value(self.value_to_enum(value))

    def _initialise_values(self):
        """Add the tango states to VALUES"""
        values_dict = {item.name: item.value for item in self.VALUES}
        values_dict.update(
            {
                "MOVING": "MOVING",
                "DISABLE": "DISABLE",
                "STANDBY": "STANDBY",
                "FAULT": "FAULT",
            }
        )
        self.VALUES = Enum("ValueEnum", values_dict)

    def get_state(self):
        """Get the device state.
        Returns:
            (enum 'HardwareObjectState'): Device state.
        """
        try:
            if self.config_values:
                _state = self.config_values[str(self.state_channel.get_value())]
            else:
                _state = str(self.state_channel.get_value())

        except (AttributeError, KeyError):
            return self.STATES.UNKNOWN

        return self.SPECIFIC_STATES[_state].value[0]

    def get_value(self):
        """Get the device value
        Returns:
            (Enum): Enum member, corresponding to the 'VALUE' or UNKNOWN.
        """
        if self.config_values:
            _val = self.config_values[str(self.state_channel.get_value())]
        else:
            _val = str(self.state_channel.get_value())
        return self.value_to_enum(_val)

    def _set_value(self, value):
        if value.name == "OPEN":
            self.open_cmd()
        elif value.name == "CLOSED":
            self.close_cmd()



from mxcubecore.BaseHardwareObjects import HardwareObject


class Mar225(HardwareObject):
    def __init__(self, name):
        super().__init__(name)

    def init(self):
        pass

    def has_shutterless(self):
        return False

    def default_mode(self):
        return 1

    def get_detector_mode(self):
        return self.default_mode()

    def set_detector_mode(self, mode):
        return



#!/usr/bin/env python
# python ednaProcLauncher.py -path
# /data/staff/biomax/staff/jie/20161212_thau_2/processed4 -mode after
# -datacollectionID 14 -residues 200 -anomalous False -cell "0,0,0,0,0,0"

import logging
import os
import string
import subprocess
import sys
import tempfile
import time
import urllib

import httplib

# XDS.INP creation is now asynchronous in mxcube, so it may not be here yet
# when we're started
WAIT_XDS_TIMEOUT = 100
HPC_HOST = "b-picard07-clu0-fe-0.maxiv.lu.se"

INPUT_TEMPLATE = """<?xml version="1.0"?>
<XSDataEDNAprocInput>
  <input_file>
    <path>
      <value>{xdsInputFile}</value>
    </path>
  </input_file>
  <data_collection_id>
    <value>{dataCollectionId}</value>
  </data_collection_id>
  <output_file>
    <path>
      <value>{output_path}</value>
    </path>
  </output_file>
{nresFragment}
{spacegroupFragment}
{cellFragment}
</XSDataEDNAprocInput>
"""

SCRIPT_TEMPLATE = """#!/usr/bin/env python

import os
import sys
import time
import socket
import traceback

sys.path.insert(0, "/mxn/groups/biomax/cmxsoft/edna-mx/kernel/src")

from EDVerbose import EDVerbose
from EDFactoryPluginStatic import EDFactoryPluginStatic

beamline = "$beamline"
proposal = "$proposal"
dataCollectionId = $dataCollectionId
ednaProcDirectory = "$ednaProcDirectory"
inputFile = "$inputFile"

pluginName = "EDPluginControlEDNAprocv1_0"
os.environ["EDNA_SITE"] = "MAXIV_BIOMAX"
os.environ["ISPyB_user"]=""
os.environ["ISPyB_pass"]=""

EDVerbose.screen("Executing EDNA plugin %s" % pluginName)
EDVerbose.screen("EDNA_SITE %s" % os.environ["EDNA_SITE"])

hostname = socket.gethostname()
dateString  = time.strftime("%Y%m%d", time.localtime(time.time()))
timeString = time.strftime("%H%M%S", time.localtime(time.time()))
#strPluginBaseDir = os.path.join("/tmp", beamline, dateString)
#if not os.path.exists(strPluginBaseDir):
#    os.makedirs(strPluginBaseDir, 0o755)

baseName = "{hostname}_{date}-{time}".format(hostname=hostname,
				 	     date=dateString,
					     time=timeString)
baseDir = os.path.join(ednaProcDirectory, baseName)
if not os.path.exists(baseDir):
    os.makedirs(baseDir, 0o755)
EDVerbose.screen("EDNA plugin working directory: %s" % baseDir)

ednaLogName = "EDNA_proc_{0}-{1}.log".format(dateString, timeString)
EDVerbose.setLogFileName(os.path.join(ednaProcDirectory, ednaLogName))
EDVerbose.setVerboseOn()

edPlugin = EDFactoryPluginStatic.loadPlugin(pluginName)
edPlugin.setDataInput(open(inputFile).read())
edPlugin.setBaseDirectory(ednaProcDirectory)
edPlugin.setBaseName(baseName)

EDVerbose.screen("Start of execution of EDNA plugin %s" % pluginName)
os.chdir(baseDir)
edPlugin.executeSynchronous()
"""


class EdnaProcLauncher:
    def __init__(
        self, path, mode, datacollectionID, residues, anomalous, cell, spacegroup
    ):
        self.autoprocessingPath = path
        self.mode = mode
        self.dataCollectionId = datacollectionID
        self.nres = residues
        self.anomalous = anomalous
        self.cell = cell
        self.spacegroup = spacegroup
        self.log = logging.getLogger()
        self.log.setLevel(logging.DEBUG)
        self.xdsAppeared = False
        self.ednaProcPath = os.path.join(self.autoprocessingPath, "EDNA_proc")
        self.ednaOutputFileName = "EDNA_proc_ispyb.xml"

        if not os.path.exists(self.ednaProcPath):
            os.makedirs(self.ednaProcPath, 0o755)
        self.xdsInputFile = os.path.join(self.autoprocessingPath, "XDS.INP")

    def parse_input_file(self):
        # parse the input file to find the first image
        self.xdsAppeared = False
        self.waitXdsStart = time.time()
        logging.getLogger("HWR").info(
            "EDNA_proc launcher: waiting for XDS.INP file: %s" % self.xdsInputFile
        )
        while not self.xdsAppeared and (
            time.time() - self.waitXdsStart < WAIT_XDS_TIMEOUT
        ):
            time.sleep(1)
            if (
                os.path.exists(self.xdsInputFile)
                and os.stat(self.xdsInputFile).st_size > 0
            ):
                time.sleep(1)
                self.xdsAppeared = True
                logging.getLogger("HWR").info(
                    "EDNA_proc launcher: XDS.INP file is there, size=%s"
                    % os.stat(self.xdsInputFile).st_size
                )
        if not self.xdsAppeared:
            logging.getLogger("HWR").error(
                "XDS.INP file ({0}) failed to appear after {1} seconds".format(
                    self.xdsInputFile, WAIT_XDS_TIMEOUT
                )
            )
        self.ednaOutputFilePath = os.path.join(
            self.ednaProcPath, self.ednaOutputFileName
        )
        if os.path.exists(self.ednaOutputFilePath):
            self.ednaOutputFile = tempfile.NamedTemporaryFile(
                suffix=".xml",
                prefix="EDNA_proc_ispyb-",
                dir=self.ednaProcPath,
                delete=False,
            )

            self.ednaOutputFilePath = os.path.join(
                self.ednaProcPath, self.ednaOutputFile.name
            )
            self.ednaOutputFile.close()
        else:
            open(self.ednaOutputFilePath, "w").write("")
        os.chmod(self.ednaOutputFilePath, 0o755)

        # ignore null nres, which might happen for whatever reason
        if self.nres is not None and self.nres != 0:
            nresFragment = """  <nres>
            <value>{0}</value>
          </nres>""".format(
                self.nres
            )
        else:
            nresFragment = ""

        if self.spacegroup is not None:
            spacegroupFragment = """  <spacegroup>
            <value>{0}</value>
          </spacegroup>""".format(
                self.spacegroup
            )
        else:
            spacegroupFragment = ""

        if self.cell is not None:
            cellFragment = """  <unit_cell>
            <value>{0}</value>
          </unit_cell>""".format(
                self.cell
            )
        else:
            cellFragment = ""

        # the other parameters are not used right now
        self.inputXml = INPUT_TEMPLATE.format(
            xdsInputFile=self.xdsInputFile,
            dataCollectionId=self.dataCollectionId,
            output_path=self.ednaOutputFilePath,
            nresFragment=nresFragment,
            cellFragment=cellFragment,
            spacegroupFragment=spacegroupFragment,
        )

        # we now need a temp file in the data dir to write the data model to
        ednaInputFileName = "EDNA_proc_input.xml"
        ednaInputFilePath = os.path.join(self.ednaProcPath, ednaInputFileName)
        if os.path.exists(ednaInputFilePath):
            # Create unique file name
            ednaInputFile = tempfile.NamedTemporaryFile(
                suffix=".xml",
                prefix="EDNA_proc_input-",
                dir=self.ednaProcPath,
                delete=False,
            )
            ednaInputFilePath = os.path.join(self.ednaProcPath, ednaInputFile.name)
            ednaInputFile.file.write(self.inputXml)
            ednaInputFile.close()
        else:
            open(ednaInputFilePath, "w").write(self.inputXml)
        os.chmod(ednaInputFilePath, 0o755)

        directories = self.autoprocessingPath.split(os.path.sep)
        try:
            beamline = directories[3]
            proposal = directories[4]
        except Exception:
            beamline = "unknown"
            proposal = "unknown"

        template = string.Template(SCRIPT_TEMPLATE)
        self.script = template.substitute(
            beamline=beamline,
            proposal=proposal,
            ednaProcDirectory=self.ednaProcPath,
            dataCollectionId=self.dataCollectionId,
            inputFile=ednaInputFilePath,
        )

        # we also need some kind of script to run edna-plugin-launcher
        ednaScriptFileName = "EDNA_proc_launcher.sh"
        self.ednaScriptFilePath = os.path.join(self.ednaProcPath, ednaScriptFileName)
        if os.path.exists(self.ednaScriptFilePath):
            # Create unique file name
            ednaScriptFile = tempfile.NamedTemporaryFile(
                suffix=".sh",
                prefix="EDNA_proc_launcher-",
                dir=self.ednaProcPath,
                delete=False,
            )
            self.ednaScriptFilePath = os.path.join(
                self.ednaProcPath, ednaScriptFile.name
            )
            ednaScriptFile.file.write(self.script)
            ednaScriptFile.close()
        else:
            open(self.ednaScriptFilePath, "w").write(self.script)
        os.chmod(self.ednaScriptFilePath, 0o755)

    def execute(self):
        cmd = (
            "echo 'cd %s;source /mxn/groups/biomax/wmxsoft/scripts_mxcube/biomax_HPC.bash_profile;/mxn/groups/biomax/cmxsoft/edna-mx/scripts_maxiv/edna_sbatch.sh %s' | ssh -F /etc/ssh/.ssh -o UserKnownHostsFile=/etc/ssh/.ssh/known_host -i /etc/ssh/id_rsa_biomax-service %s;source /mxn/groups/biomax/wmxsoft/scripts_mxcube/biomax_HPC.bash_profile"
            % (self.ednaProcPath, self.ednaScriptFilePath, HPC_HOST)
        )

        # for test
        # cmd = "echo 'cd %s;/mxn/groups/biomax/cmxsoft/edna-mx/scripts_maxiv/edna_sbatch.sh %s' | ssh %s" % (autoPROCPath, ednaScriptFilePath, hpc_host)
        # print cmd
        logging.getLogger("HWR").info(
            "EDNA_proc launcher: command gonna be launched: %s" % cmd
        )
        p = subprocess.Popen(
            cmd, shell=True
        )  # , stdout = subprocess.PIPE, stderr = subprocess.PIPE)
        p.wait()

    def parse_and_execute(self):
        self.parse_input_file()
        self.execute()


if __name__ == "__main__":
    args = sys.argv[1:]

    if (len(args) % 2) != 0:
        logging.error(
            "the argument list is not well formed (odd number of args/options)"
        )
        sys.exit()

    # do the arg parsing by hand since neither getopt nor optparse support
    # single dash long options.
    options = dict()
    for x in range(0, len(args), 2):
        options[args[x]] = args[x + 1]

    autoprocessingPath = options["-path"]

    residues = float(options.get("-residues", 200))
    anomalous = options.get("-anomalous", False)
    spacegroup = options.get("-sg", None)
    cell = options.get("-cell", "0,0,0,0,0,0")
    dataCollectionId = options["-datacollectionID"]
    mode = options.get("-mode")

    edna = EdnaProcLauncher(
        autoprocessingPath,
        mode,
        dataCollectionId,
        residues,
        anomalous,
        cell,
        spacegroup,
    )
    edna.parse_and_execute()



#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""
Hardware Object is used to get relevant machine information
(current, intensity, hutch temperature and humidity, and data storage disc
information). Value limits are included
"""

import logging
import os
import time

try:
    from urllib2 import urlopen
except ImportError:
    from urllib.request import urlopen

from collections import OrderedDict
from datetime import (
    datetime,
    timedelta,
)

from gevent import spawn

from mxcubecore import HardwareRepository as HWR
from mxcubecore.BaseHardwareObjects import HardwareObject

__credits__ = ["EMBL Hamburg"]
__license__ = "LGPLv3+"
__category__ = "General"


class EMBLMachineInfo(HardwareObject):
    """Displays actual information about the beamline
    """

    def __init__(self, name):
        """OrderedDict is used to have a sorted items for display
           and directory like access when updating values
        """

        HardwareObject.__init__(self, name)

        self.update_interval = None
        self.limits_dict = None
        self.hutch_temp_addr = None
        self.hutch_hum_addr = None
        self.hutch_temp = 0
        self.hutch_hum = 0
        self.overflow_alarm = None
        self.low_level_alarm = None
        self.state_text = ""
        self.ring_energy = None
        self.bunch_count = None
        self.flux_area = None
        self.last_transmission = None
        self.frontend_is_open = False
        self.undulator_gap = 9999

        self.values_ordered_dict = OrderedDict()

        self.values_ordered_dict["current"] = {
            "value": 0,
            "value_str": "",
            "in_range": False,
            "title": "Machine current",
            "bold": True,
        }

        self.values_ordered_dict["machine_state"] = {
            "value": None,
            "in_range": False,
            "title": "Machine state",
        }

        self.values_ordered_dict["frontend_undulator"] = {
            "value": None,
            "in_range": True,
            "title": "Front End, undulator gap",
        }

        self.values_ordered_dict["temp_hum"] = {
            "value": "",
            "value_str": "",
            "in_range": None,
            "title": "Hutch temperature and humidity",
        }

        self.temp_hum_values = [None, None]
        self.temp_hum_in_range = [None, None]
        self.temp_hum_polling = None

        self.chan_mach_curr = None
        self.chan_mach_energy = None
        self.chan_bunch_count = None
        self.chan_frontend_status = None
        self.chan_undulator_gap = None
        self.chan_state_text = None
        self.chan_cryojet_in = None
        self.chan_sc_dewar_low_level_alarm = None
        self.chan_sc_dewar_overflow_alarm = None

    def init(self):

        self.update_interval = int(self.get_property("updateIntervalS"))
        self.limits_dict = eval(self.get_property("limits"))
        self.hutch_temp_addr = self.get_property("hutchTempAddress")
        self.hutch_hum_addr = self.get_property("hutchHumAddress")

        self.chan_mach_curr = self.get_channel_object("machCurrent")
        self.chan_mach_curr.connect_signal("update", self.mach_current_changed)
        self.chan_state_text = self.get_channel_object("machStateText")
        self.chan_state_text.connect_signal("update", self.state_text_changed)
        # self.state_text_changed(self.chan_state_text.get_value())

        self.chan_mach_energy = self.get_channel_object("machEnergy")
        self.chan_mach_energy.connect_signal("update", self.mach_energy_changed)
        self.chan_bunch_count = self.get_channel_object("machBunchCount")
        self.chan_bunch_count.connect_signal("update", self.bunch_count_changed)
        self.chan_frontend_status = self.get_channel_object("frontEndStatus")
        self.chan_frontend_status.connect_signal("update", self.frontend_status_changed)
        # self.frontend_status_changed(self.chan_frontend_status.get_value())

        if HWR.beamline.flux is not None:
            self.connect(HWR.beamline.flux, "fluxInfoChanged", self.flux_info_changed)
            self.values_ordered_dict["flux"] = {
                "value": 1,
                "value_str": "Remeasure flux!",
                "in_range": False,
                "title": "Measured flux",
            }

        self.chan_undulator_gap = self.get_channel_object("chanUndulatorGap")
        if self.chan_undulator_gap is not None:
            self.chan_undulator_gap.connect_signal("update", self.undulator_gap_changed)
        self.undulator_gap_changed(self.chan_undulator_gap.get_value())

        self.chan_cryojet_in = self.get_channel_object("cryojetIn", optional=True)
        if self.chan_cryojet_in is not None:
            self.values_ordered_dict["cryo"] = {
                "value": "???",
                "in_range": None,
                "title": "Cryoject in place",
            }
            self.cryojet_in_changed(self.chan_cryojet_in.get_value())
            self.chan_cryojet_in.connect_signal("update", self.cryojet_in_changed)
        else:
            logging.getLogger("HWR").debug("MachineInfo: Cryojet channel not defined")

        self.chan_sc_dewar_low_level_alarm = self.get_channel_object(
            "scLowLevelAlarm", optional=True
        )
        if self.chan_sc_dewar_low_level_alarm is not None:
            self.values_ordered_dict["sc"] = {
                "value": "Dewar level in range",
                "in_range": True,
                "title": "Sample changer",
            }

            self.chan_sc_dewar_low_level_alarm.connect_signal(
                "update", self.low_level_alarm_changed
            )
            self.low_level_alarm_changed(self.chan_sc_dewar_low_level_alarm.get_value())

        self.chan_sc_dewar_overflow_alarm = self.get_channel_object(
            "scOverflowAlarm", optional=True
        )
        if self.chan_sc_dewar_overflow_alarm is not None:
            self.chan_sc_dewar_overflow_alarm.connect_signal(
                "update", self.overflow_alarm_changed
            )

        if hasattr(HWR.beamline, "ppu_control"):
            self.values_ordered_dict["ppu"] = {
                "value": "- - -",
                "in_range": False,
                "title": "Files copied - pending - failed",
            }

            self.connect(
                HWR.beamline.ppu_control,
                "fileTranferStatusChanged",
                self.file_transfer_status_changed,
            )

        self.chan_count_dropped = self.get_channel_object(
            "framesCountDropped", optional=True
        )
        if self.chan_count_dropped is not None:
            self.values_ordered_dict["frames_dropped"] = {
                "value": "",
                "in_range": True,
                "title": "Frames dropped",
            }
            self.chan_count_dropped.connect_signal("update", self.count_dropped_changed)

        self.temp_hum_polling = spawn(
            self.get_temp_hum_values, self.get_property("updateIntervalS")
        )

        self.re_emit_values()

    def clear_gevent(self):
        """Clear gevent tasks

        :return: None
        """
        self.temp_hum_polling.kill()
        if self.update_task:
            self.update_task.kill()

    def cryojet_in_changed(self, value):
        """Updates cryojet status

        :param value: status
        :type value: bool
        :return: None
        """

        self.values_ordered_dict["cryo"]["in_range"] = False
        self.values_ordered_dict["cryo"]["bold"] = True

        if value == 1:
            self.values_ordered_dict["cryo"]["value"] = " In place"
            self.values_ordered_dict["cryo"]["in_range"] = True
            self.values_ordered_dict["cryo"]["bold"] = False
        elif value == 0:
            self.values_ordered_dict["cryo"]["value"] = "NOT IN PLACE"
        else:
            self.values_ordered_dict["cryo"]["value"] = "Unknown"
        self.re_emit_values()

    def mach_current_changed(self, value):
        """Method called if the machine current is changed

        :param value: new machine current
        :type value: float
        """
        if (
            self.values_ordered_dict["current"]["value"] is None
            or abs(self.values_ordered_dict["current"]["value"] - value) > 0.00001
        ):
            self.values_ordered_dict["current"]["value"] = value
            self.values_ordered_dict["current"]["value_str"] = "%.1f mA" % value
            self.values_ordered_dict["current"]["in_range"] = value > 60.0
            self.re_emit_values()

    def state_text_changed(self, text):
        """Function called if machine state text is changed

        :param text: new machine state text
        :type text: string
        """
        self.state_text = str(text)
        self.values_ordered_dict["machine_state"]["in_range"] = text != "Fehler"
        self.update_machine_state()

    def mach_energy_changed(self, value):
        """Updates machine energy value

        :param value: machine energy
        :type value: float
        :return: None
        """
        self.ring_energy = value
        self.update_machine_state()

    def bunch_count_changed(self, value):
        """Bunch count changed"""
        self.bunch_count = value
        self.update_machine_state()

    def frontend_status_changed(self, value):
        """
        Update front end status
        :param value:
        :return:
        """
        self.frontend_is_open = value[2] == 2
        self.update_machine_state()

    def undulator_gap_changed(self, value):
        """
        Update undulator gaps
        :param value: float
        :return:
        """
        if isinstance(value, (list, tuple)):
            value = value[0]
        self.undulator_gap = value / 1000

    def update_machine_state(self):
        """Machine state assembly"""
        state_text = self.state_text
        if self.ring_energy is not None:
            state_text += "\n%.2f GeV " % self.ring_energy
        if self.bunch_count is not None:
            state_text += ", %d Bunches" % self.bunch_count
        self.values_ordered_dict["machine_state"]["value"] = state_text

        if not self.frontend_is_open or self.undulator_gap > 30:
            self.values_ordered_dict["frontend_undulator"]["in_range"] = False
        else:
            self.values_ordered_dict["frontend_undulator"]["in_range"] = True
        if self.frontend_is_open:
            self.values_ordered_dict["frontend_undulator"]["value_str"] = (
                "Opened, %d mm" % self.undulator_gap
            )
        else:
            self.values_ordered_dict["frontend_undulator"]["value_str"] = (
                "Closed, %d mm" % self.undulator_gap
            )

        self.re_emit_values()

    def low_level_alarm_changed(self, value):
        """Low level alarm"""
        self.low_level_alarm = value
        self.update_sc_alarm()

    def overflow_alarm_changed(self, value):
        """Overflow alarm"""
        self.overflow_alarm = value
        self.update_sc_alarm()

    def file_transfer_status_changed(self, status):
        """
        Updates info about file beeing transfered
        :param total: int
        :param pending: int
        :param failed: int
        :return:
        """
        self.values_ordered_dict["ppu"]["value"] = "%d  -  %d  -  %d" % (
            status[0],
            status[1],
            status[2],
        )
        self.values_ordered_dict["ppu"]["in_range"] = status[2] == 0

        if status[2] > 0:
            logging.getLogger("GUI").error(
                "Error in file transfer (%d files failed to copy)." % status[2]
            )

    def count_dropped_changed(self, num_dropped):
        self.values_ordered_dict["frames_dropped"]["value"] = str(num_dropped)
        self.values_ordered_dict["frames_dropped"]["in_range"] = num_dropped == 0
        if num_dropped > 0:
            logging.getLogger("GUI").error(
                "Error during the frame acquisition (in total %d frame(s) dropped)."
                % num_dropped
            )

    def update_sc_alarm(self):
        """Sample changer alarm"""
        if self.low_level_alarm == 1:
            self.values_ordered_dict["sc"]["value"] = "Low level alarm!"
            self.values_ordered_dict["sc"]["in_range"] = False
            self.values_ordered_dict["sc"]["bold"] = True
            logging.getLogger("GUI").error(
                "Liquid nitrogen level in sample changer dewar is too low!"
            )

        elif self.overflow_alarm:
            self.values_ordered_dict["sc"]["value"] = "Overflow alarm!"
            self.values_ordered_dict["sc"]["in_range"] = False
            self.values_ordered_dict["sc"]["bold"] = True
            logging.getLogger("GUI").error(
                "Liquid nitrogen overflow in sample changer dewar!"
            )
        else:
            self.values_ordered_dict["sc"]["value"] = "Dewar level in range"
            self.values_ordered_dict["sc"]["in_range"] = True
        self.re_emit_values()

    def flux_info_changed(self, flux_info):
        """Sets flux value"""

        if flux_info["measured"] is None:
            self.values_ordered_dict["flux"]["value"] = 0
            self.values_ordered_dict["flux"][
                "value_str"
            ] = "Beamline mode changed\nRemeasure flux!"
            self.values_ordered_dict["flux"]["in_range"] = False
        else:
            msg_str = "Flux: %.2E ph/s\n" % flux_info["measured"]["flux"]
            msg_str += "%d%% transmission, %dx%d beam" % (
                flux_info["measured"]["transmission"],
                flux_info["measured"]["size_x"] * 1000,
                flux_info["measured"]["size_y"] * 1000,
            )

            self.values_ordered_dict["flux"]["value"] = flux_info["measured"]["flux"]
            self.values_ordered_dict["flux"]["value_str"] = msg_str
            self.values_ordered_dict["flux"]["in_range"] = (
                flux_info["measured"]["flux"] > 1e6
            )
        self.re_emit_values()

    def re_emit_values(self):
        """Emits list of values"""
        self.emit("valuesChanged", self.values_ordered_dict)

    def get_values(self):
        """Returns list of values"""
        return self.values_ordered_dict

    def get_temp_hum_values(self, sleep_time):
        """Updates temperatur and humidity values"""
        while True:
            temp = self.get_external_value(self.hutch_temp_addr)
            hum = self.get_external_value(self.hutch_hum_addr)
            if not None in (temp, hum):
                if abs(float(temp) - self.hutch_temp) > 0.1 or abs(
                    float(hum) != self.hutch_hum > 1
                ):
                    self.hutch_temp = temp
                    self.hutch_hum = hum
                    self.values_ordered_dict["temp_hum"][
                        "value"
                    ] = "%.1f C, %.1f %%" % (temp, hum)
                    self.values_ordered_dict["temp_hum"]["in_range"] = (
                        temp < 25 and hum < 60
                    )
                    self.re_emit_values()
            time.sleep(sleep_time)

    def get_current(self):
        """Returns machine current in mA
        """
        return self.values_ordered_dict["current"]["value"]

    def get_current_value(self):
        """Returns machine current in mA"""
        return self.values_ordered_dict["current"]["value"]

    def get_message(self):
        """Returns synchrotron state text"""
        return self.state_text

    def get_external_value(self, addr):
        """Extracts value from the given epics address. This is very specific
           implementation how to get a value from epics web tool. At first
           web address string is formed and then web page by urllib2
           extracted. Page contains column with records.
           Then the last value is choosen as the last active value.

        :param addr: epics address
        :type addr: str
        :returns : float
        """
        url_prefix = (
            "http://cssweb.desy.de:8084/ArchiveViewer/archive"
            + "reader.jsp?DIRECTORY=%2Fdata7%2FChannelArchiver%"
            + "2FchannelReference2.kryo&PATTERN=&"
        )
        end = datetime.now()
        start = end - timedelta(hours=24)
        url_date = "=on&STARTMONTH=%d&STARTDAY=%d&STARTYEAR=%d" % (
            start.month,
            start.day,
            start.year,
        ) + "&STARTHOUR=%d&STARTMINUTE=%d&STARTSECOND=0" % (start.hour, start.minute)
        url_date = url_date + (
            "&ENDMONTH=%d&ENDDAY=%d&ENDYEAR=%d" % (end.month, end.day, end.year)
            + "&ENDHOUR=%d&ENDMINUTE=%d&ENDSECOND=0" % (end.hour, (end.minute - 10))
        )
        url_date = (
            url_date
            + "&COMMAND=GET&Y0=0&Y1=0&FORMAT=SPREADSHEET&"
            + "INTERPOL=0&NUMOFPOINTS=10"
        )
        url_file = None
        last_value = None
        try:
            addr = addr.split(":")
            url_device = (
                "NAMES=" + addr[0] + "%3A" + addr[1] + "%3A" + addr[2] + "%3A" + addr[3]
            )
            url_device = (
                url_device
                + "&FRAME2=1"
                + addr[0]
                + "%3A"
                + addr[1]
                + "%3A"
                + addr[2]
                + "%3A"
                + addr[3]
            )
            url_device = (
                url_device
                + "&NAMES2=&"
                + addr[0]
                + "%3A"
                + addr[1]
                + "%3A"
                + addr[2]
                + "%3A"
                + addr[3]
            )
            final_url = url_prefix + url_device + url_date
            url_file = urlopen(final_url)
            for line in url_file:
                line_el = line.split()
                if line_el:
                    if line_el[-1].isdigit:
                        last_value = line_el[-1]
            last_value = float(last_value)
        except Exception:
            logging.getLogger("HWR").debug("MachineInfo: Unable to read epics values")
        finally:
            if url_file:
                url_file.close()
        return last_value

    def get_ramdisk_size(self):
        """
        Gets ramdisk size in bytes
        :return: total, free disc size in bytes and free disc in perc
        """
        data_dir = "/ramdisk/"
        # p = '/' + data_dir.split('/')[1]
        # data_dir = str(p)
        if os.path.exists(data_dir):
            st = os.statvfs(data_dir)

            total = st.f_blocks * st.f_frsize
            free = st.f_bavail * st.f_frsize
            perc = st.f_bavail / float(st.f_blocks)
            return total, free, perc
        else:
            return None, None, None

    def sizeof_fmt(self, num):
        """Returns disk space formated in string"""

        try:
            for x in ["bytes", "KB", "MB", "GB"]:
                if num < 1024.0:
                    return "%3.1f%s" % (num, x)
                num /= 1024.0
            return "%3.1f%s" % (num, "TB")
        except Exception:
            return "???"



# encoding: utf-8
#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

__copyright__ = """Copyright The MXCuBE Collaboration"""
__license__ = "LGPLv3+"

from collections import OrderedDict

from mxcubecore.BaseHardwareObjects import HardwareObjectState
from mxcubecore.HardwareObjects.abstract.AbstractActuator import AbstractActuator
from mxcubecore.HardwareObjects.abstract.AbstractMotor import MotorStates


class MotorsNPosition(AbstractActuator):
    """
    <object class="MotorsNState">
        <username>Dummy</username>

        <motors>
            <object role="role1" href="/mot1"></object>
            <object role="role2" href="/mot2"></object>
        </motors>

        <deltas>
            <role1>0.1</role1>
            <role2>0.3</role1>
        </deltas>

        <positions>
            <position>
                <name>BEAM</name>
                <role1>0.12</role1>
                <role2>0.22</role2>
                <favproperty>I am in the Beam</favproperty>
            </position>

            <position>
                <name>OUTBEAM</name>
                <role1>2.3</role1>
                <role2>4.6</role2>
                <favproperty>I am out of Beam</favproperty>
            </position>
        </positions>

    </object>
    """

    default_delta = 0.01

    def __init__(self, name):
        AbstractActuator.__init__(self, name)

        self.motor_hwobjs = {}
        self.deltas = {}

        self._positions = OrderedDict()
        self._properties = OrderedDict()
        self.motorlist = []

        self.current_index = None

        self._last_position_name = None
        self._updating_multi_value = None

    def init(self):
        motorlist = self.get_property("motorlist").split(",")
        self.motorlist = [motor.strip() for motor in motorlist]

        for motorname in self.motorlist:
            motor_hwo = self.get_object_by_role(motorname)
            self.motor_hwobjs[motorname] = motor_hwo

            delta = self["deltas"].get_property(motorname, self.default_delta)
            self.deltas[motorname] = delta

            self.connect(motor_hwo, "stateChanged", self.motor_state_changed)
            self.connect(motor_hwo, "valueChanged", self.motor_value_changed)

        self.load_positions()

        self.update_multi_value()
        self.update_state(HardwareObjectState.READY)

    def load_positions(self):
        positions = self["positions"]

        for position in positions:
            name = position.get_property("name")

            self._positions[name] = {}
            self._properties[name] = {}

            for motorname in self.motorlist:
                pos = position.get_property(motorname)
                self._positions[name][motorname] = pos

            for prop in position.get_properties():
                if prop not in self.motorlist:
                    if prop != "name":
                        value = position.get_property(prop)
                        self._properties[name][prop] = value

    def get_position_list(self):
        return list(self._positions.keys())

    def get_property_value_by_index(self, position_index=None, property_name=None):
        """
        Returns the value of the property with the given property_name at the specified position_index.
        If the property_name or position_index is invalid, returns None.
        """
        if position_index is not None and 0 <= position_index < len(self._positions):
            position_name = list(self._positions.keys())[position_index]
            return self._properties.get(position_name, {}).get(property_name)
        return None

    def get_value(self):
        return self.update_multi_value()

    def set_position(self, posname):
        """
        Allow to move by providing in order (if previous not found):
               posname
               name
               index
        """

        # by posname - alias
        for posidx, name in enumerate(self._positions):
            if posname == self.get_property_value_by_index(posidx, "posname"):
                self._set_value(posidx)
                return

        # by name - label
        for posidx, name in enumerate(self._positions):
            if posname == name:
                self._set_value(posidx)
                return

        if isinstance(posname, int):
            if posname >= 0 and posname < len(self._positions):
                self._set_value(posname)
                return

        self.user_log.error(
            "Wrong position name %s selected for %s" % (posname, self.username)
        )

    def _set_value(self, value):
        if value >= 0 and value < len(self._positions):
            name = list(self._positions.keys())[value]
            for motorname in self.motorlist:
                pos = self._positions[name][motorname]
                motor_hwobj = self.motor_hwobjs[motorname]
                motor_hwobj.set_value(pos)

    def get_position(self):
        current_idx = self.get_value()
        if current_idx is not None and current_idx != -1:
            return list(self._positions.keys())[current_idx]

    def motor_state_changed(self, state):
        self.update_multi_state()

    def motor_value_changed(self, position=None):
        self.update_multi_value()

    def update_multi_value(self):
        current_idx = -1

        if not self._updating_multi_value:
            self._updating_multi_value = True

            try:
                current_pos = {
                    motorname: self.motor_hwobjs[motorname].get_value()
                    for motorname in self.motorlist
                }

                for idx, name in enumerate(self._positions):
                    for motorname in self.motorlist:
                        if motorname not in self._positions[name]:
                            continue

                        position = self._positions[name][motorname]
                        cur_pos = current_pos[motorname]
                        delta = self.deltas[motorname]

                        if abs(cur_pos - position) > delta:
                            break
                    else:
                        for motorname in self.motorlist:
                            position = self._positions[name][motorname]
                            self.log.debug(
                                "     - motor %s is at %s" % (motorname, position)
                            )
                        current_idx = idx
                        break

                if current_idx != self.current_index:
                    self.current_index = current_idx
                    self.update_value(current_idx)

            finally:
                self._updating_multi_value = False

        return current_idx

    def update_multi_state(self):

        multi_state = HardwareObjectState.READY

        for motorname in self.motor_hwobjs:
            motor = self.motor_hwobjs[motorname]
            state = motor.get_state()

            self.log.debug(
                "MotorsNPosition - updating multi_state. motor (%s) is %s"
                % (motorname, str(state))
            )

            if state in (HardwareObjectState.FAULT, HardwareObjectState.UNKNOWN):
                multi_state = state
                break

            if state == HardwareObjectState.OFF:
                multi_state = state
                continue

            if state in (MotorStates.MOVING, HardwareObjectState.BUSY):
                if multi_state != HardwareObjectState.OFF:
                    multi_state = HardwareObjectState.BUSY
                    continue

        self.update_state(multi_state)

    def stop(self):
        """
        Descript. :
        """
        for motor in self.motor_hwobjs:
            motor.stop()

    def is_moving(self):
        """
        Descript. : True if the motor is currently moving
        """
        return self.get_state() == HardwareObjectState.BUSY



EDNA_TEST_DATA = """<?xml version="1.0" ?>
<XSDataInputMXCuBE>
  <dataCollectionId>
    <value>976533</value>
  </dataCollectionId>
  <diffractionPlan>
    <aimedIOverSigmaAtHighestResolution>
      <value>3.000000e+00</value>
    </aimedIOverSigmaAtHighestResolution>
    <complexity>
      <value>none</value>
    </complexity>
    <maxExposureTimePerDataCollection>
      <value>6.000000e+03</value>
    </maxExposureTimePerDataCollection>
  </diffractionPlan>
  <experimentalCondition>
    <beam>
      <flux>
        <value>2.710000e+12</value>
      </flux>
      <minExposureTimePerImage>
        <value>8.000000e-02</value>
      </minExposureTimePerImage>
      <size>
        <x>
          <value>5.000000e-02</value>
        </x>
        <y>
          <value>3.000000e-02</value>
        </y>
      </size>
      <transmission>
        <value>1.000000e+02</value>
      </transmission>
      <wavelength>
        <value>9.136656e-01</value>
      </wavelength>
    </beam>
    <goniostat>
      <maxOscillationSpeed>
        <value>3.600000e+02</value>
      </maxOscillationSpeed>
      <minOscillationWidth>
        <value>5.000000e-02</value>
      </minOscillationWidth>
    </goniostat>
  </experimentalCondition>
  <sample>
    <shape>
      <value>1.000000e+00</value>
    </shape>
    <size>
      <x>
        <value>1.000000e-01</value>
      </x>
      <y>
        <value>1.000000e-01</value>
      </y>
      <z>
        <value>1.000000e-01</value>
      </z>
    </size>
    <susceptibility>
      <value>1.000000e+00</value>
    </susceptibility>
  </sample>
  <dataSet>
    <imageFile>
      <path>
        <value>/data/id29/inhouse/DAS/RAW_DATA/ref-4ESR372D11_1_0001.cbf</value>
      </path>
    </imageFile>
    <imageFile>
      <path>
        <value>/data/id29/inhouse/DAS/RAW_DATA/ref-4ESR372D11_1_0002.cbf</value>
      </path>
    </imageFile>
    <imageFile>
      <path>
        <value>/data/id29/inhouse/DAS/RAW_DATA/ref-4ESR372D11_1_0003.cbf</value>
      </path>
    </imageFile>
    <imageFile>
      <path>
        <value>/data/id29/inhouse/DAS/RAW_DATA/ref-4ESR372D11_1_0004.cbf</value>
      </path>
    </imageFile>
  </dataSet>
</XSDataInputMXCuBE>
"""

EDNA_DEFAULT_INPUT = """<?xml version="1.0" ?>
<XSDataInputMXCuBE>
    <diffractionPlan>
        <aimedCompleteness>
            <value>9.900000e-01</value>
        </aimedCompleteness>
        <aimedIOverSigmaAtHighestResolution>
            <value>3.000000e+00</value>
        </aimedIOverSigmaAtHighestResolution>
        <aimedResolution>
            <value>1.000000e+00</value>
        </aimedResolution>
        <aimedMultiplicity>
            <value>4.000000e+00</value>
        </aimedMultiplicity>
        <complexity>
            <value>min</value>
        </complexity>
        <maxExposureTimePerDataCollection>
            <value>6.000000e+03</value>
        </maxExposureTimePerDataCollection>
    </diffractionPlan>
    <experimentalCondition>
        <beam>
            <minExposureTimePerImage>
                <value>1.000000e-01</value>
            </minExposureTimePerImage>
            <size>
                <x>
                    <value>1.000000e-01</value>
                </x>
                <y>
                    <value>0.70000e-01</value>
                </y>
            </size>
        </beam>
        <goniostat>
            <maxOscillationSpeed>
                <value>3.600000e+02</value>
            </maxOscillationSpeed>
            <minOscillationWidth>
                <value>2.000000e-01</value>
            </minOscillationWidth>
        </goniostat>
    </experimentalCondition>
    <sample>
        <size>
            <x>
                <value>1.000000e-01</value>
            </x>
            <y>
                <value>1.000000e-01</value>
            </y>
            <z>
                <value>1.000000e-01</value>
            </z>
        </size>
        <susceptibility>
            <value>1.000000e+00</value>
        </susceptibility>
    </sample>
</XSDataInputMXCuBE>
"""

EDNA_RESULT_DATA = """<?xml version="1.0" ?>
<XSDataResultMXCuBE>
    <characterisationResult>
        <dataCollection>
            <diffractionPlan>
                <aimedCompleteness>
                    <value>9.900000e-01</value>
                </aimedCompleteness>
                <aimedIOverSigmaAtHighestResolution>
                    <value>2.000000e+00</value>
                </aimedIOverSigmaAtHighestResolution>
                <aimedResolution>
                    <value>5.000000e-01</value>
                </aimedResolution>
                <anomalousData>
                    <value>false</value>
                </anomalousData>
                <complexity>
                    <value>none</value>
                </complexity>
                <detectorDistanceMax>
                    <value>1.300000e+03</value>
                </detectorDistanceMax>
                <detectorDistanceMin>
                    <value>1.350000e+02</value>
                </detectorDistanceMin>
                <forcedSpaceGroup>
                    <value></value>
                </forcedSpaceGroup>
                <maxExposureTimePerDataCollection>
                    <value>6.000000e+02</value>
                </maxExposureTimePerDataCollection>
            </diffractionPlan>
            <subWedge>
                <experimentalCondition>
                    <beam>
                        <exposureTime>
                            <value>4.000000e-02</value>
                        </exposureTime>
                        <flux>
                            <value>2.300000e+12</value>
                        </flux>
                        <minExposureTimePerImage>
                            <value>4.000000e-02</value>
                        </minExposureTimePerImage>
                        <size>
                            <x>
                                <value>5.000000e-02</value>
                            </x>
                            <y>
                                <value>5.000000e-02</value>
                            </y>
                        </size>
                        <transmission>
                            <value>1.000000e+02</value>
                        </transmission>
                        <wavelength>
                            <value>9.762460e-01</value>
                        </wavelength>
                    </beam>
                    <detector>
                        <beamPositionX>
                            <value>2.159546e+02</value>
                        </beamPositionX>
                        <beamPositionY>
                            <value>2.103680e+02</value>
                        </beamPositionY>
                        <distance>
                            <value>3.920700e+02</value>
                        </distance>
                        <name>
                            <value>PILATUS 6M</value>
                        </name>
                        <numberPixelX>
                            <value>2463</value>
                        </numberPixelX>
                        <numberPixelY>
                            <value>2527</value>
                        </numberPixelY>
                        <pixelSizeX>
                            <value>1.720000e-01</value>
                        </pixelSizeX>
                        <pixelSizeY>
                            <value>1.720000e-01</value>
                        </pixelSizeY>
                        <serialNumber>
                            <value>PILATUS 6M-F, S/N 60-0117-F</value>
                        </serialNumber>
                        <type>
                            <value>pilatus6m</value>
                        </type>
                    </detector>
                    <goniostat>
                        <maxOscillationSpeed>
                            <value>1.190000e+02</value>
                        </maxOscillationSpeed>
                        <minOscillationWidth>
                            <value>5.000000e-02</value>
                        </minOscillationWidth>
                        <oscillationWidth>
                            <value>1.000000e+00</value>
                        </oscillationWidth>
                        <rotationAxisEnd>
                            <value>9.100000e+01</value>
                        </rotationAxisEnd>
                        <rotationAxisStart>
                            <value>9.000000e+01</value>
                        </rotationAxisStart>
                    </goniostat>
                </experimentalCondition>
                <subWedgeNumber>
                    <value>1</value>
                </subWedgeNumber>
            </subWedge>
            <subWedge>
                <experimentalCondition>
                    <beam>
                        <exposureTime>
                            <value>4.000000e-02</value>
                        </exposureTime>
                        <flux>
                            <value>2.300000e+12</value>
                        </flux>
                        <minExposureTimePerImage>
                            <value>4.000000e-02</value>
                        </minExposureTimePerImage>
                        <size>
                            <x>
                                <value>5.000000e-02</value>
                            </x>
                            <y>
                                <value>5.000000e-02</value>
                            </y>
                        </size>
                        <transmission>
                            <value>1.000000e+02</value>
                        </transmission>
                        <wavelength>
                            <value>9.762460e-01</value>
                        </wavelength>
                    </beam>
                    <goniostat>
                        <maxOscillationSpeed>
                            <value>1.190000e+02</value>
                        </maxOscillationSpeed>
                        <minOscillationWidth>
                            <value>5.000000e-02</value>
                        </minOscillationWidth>
                        <oscillationWidth>
                            <value>1.000000e+00</value>
                        </oscillationWidth>
                        <rotationAxisEnd>
                            <value>1.810000e+02</value>
                        </rotationAxisEnd>
                        <rotationAxisStart>
                            <value>1.800000e+02</value>
                        </rotationAxisStart>
                    </goniostat>
                </experimentalCondition>
                <image>
                    <path>
                    </path>
                    <number>
                        <value>2</value>
                    </number>
                </image>
                <subWedgeNumber>
                    <value>2</value>
                </subWedgeNumber>
            </subWedge>
        </dataCollection>
        <strategyResult>
            <collectionPlan>
                <collectionPlanNumber>
                    <value>1</value>
                </collectionPlanNumber>
                <collectionStrategy>
                    <subWedge>
                        <experimentalCondition>
                            <beam>
                                <exposureTime>
                                    <value>4.000000e-02</value>
                                </exposureTime>
                                <flux>
                                    <value>2.300000e+12</value>
                                </flux>
                                <minExposureTimePerImage>
                                    <value>4.000000e-02</value>
                                </minExposureTimePerImage>
                                <size>
                                    <x>
                                        <value>5.000000e-02</value>
                                    </x>
                                    <y>
                                        <value>5.000000e-02</value>
                                    </y>
                                </size>
                                <transmission>
                                    <value>3.136190e+01</value>
                                </transmission>
                                <wavelength>
                                    <value>9.762460e-01</value>
                                </wavelength>
                            </beam>
                            <detector>
                                <beamPositionX>
                                    <value>2.156012e+02</value>
                                </beamPositionX>
                                <beamPositionY>
                                    <value>2.115154e+02</value>
                                </beamPositionY>
                                <distance>
                                    <value>3.593500e+02</value>
                                </distance>
                                <name>
                                    <value>PILATUS 6M</value>
                                </name>
                                <numberPixelX>
                                    <value>2463</value>
                                </numberPixelX>
                                <numberPixelY>
                                    <value>2527</value>
                                </numberPixelY>
                                <pixelSizeX>
                                    <value>1.720000e-01</value>
                                </pixelSizeX>
                                <pixelSizeY>
                                    <value>1.720000e-01</value>
                                </pixelSizeY>
                                <serialNumber>
                                    <value>PILATUS 6M-F</value>
                                </serialNumber>
                                <type>
                                    <value>pilatus6m</value>
                                </type>
                            </detector>
                            <goniostat>
                                <maxOscillationSpeed>
                                    <value>1.190000e+02</value>
                                </maxOscillationSpeed>
                                <minOscillationWidth>
                                    <value>5.000000e-02</value>
                                </minOscillationWidth>
                                <oscillationWidth>
                                    <value>1.000000e-01</value>
                                </oscillationWidth>
                                <rotationAxisEnd>
                                    <value>2.420000e+02</value>
                                </rotationAxisEnd>
                                <rotationAxisStart>
                                    <value>1.380000e+02</value>
                                </rotationAxisStart>
                            </goniostat>
                        </experimentalCondition>
                        <subWedgeNumber>
                            <value>1</value>
                        </subWedgeNumber>
                    </subWedge>
                </collectionStrategy>
                <strategySummary>
                    <completeness>
                        <value>1.000000e+00</value>
                    </completeness>
                    <iSigma>
                        <value>2.000000e+00</value>
                    </iSigma>
                    <rankingResolution>
                        <value>1.810000e+00</value>
                    </rankingResolution>
                    <redundancy>
                        <value>4.440000e+00</value>
                    </redundancy>
                    <resolution>
                        <value>1.810000e+00</value>
                    </resolution>
                    <resolutionReasoning>
                        <value>Resolution limit is set by the radiation damage</value>
                    </resolutionReasoning>
                    <totalDataCollectionTime>
                        <value>4.409600e+01</value>
                    </totalDataCollectionTime>
                    <totalExposureTime>
                        <value>72</value>
                    </totalExposureTime>
                </strategySummary>
            </collectionPlan>
        </strategyResult>
    </characterisationResult>
    <htmlPage>
        <path>
        </path>
    </htmlPage>
    <screeningId>
        <value>21514</value>
    </screeningId>
</XSDataResultMXCuBE>
"""



# -*- coding: utf-8 -*-
"""
[Name] BeamInfo

[Description]
BeamInfo hardware object informs mxCuBE (HutchMenuBrick) about the beam position
and size.

This is the Soleil PX1 version

[Emited signals]

beamInfoChanged
beamPosChanged

[Included Hardware Objects]

[Example XML file]

<object class = "BeaminfoPX2">
  <username>Beamstop</username>
  <channel type="tango" tangoname="i11-ma-cx1/ex/md2" polling="1000" name="beamsizex">BeamSizeHorizontal</channel>
  <channel type="tango" tangoname="i11-ma-cx1/ex/md2" polling="1000" name="beamsizey">BeamSizeVertical</channel>
  <channel type="tango" tangoname="i11-ma-cx1/ex/md2" polling="1000" name="positionx">BeamPositionHorizontal</channel>
  <channel type="tango" tangoname="i11-ma-cx1/ex/md2" polling="1000" name="positiony">BeamPositionVertical</channel>
  <object  role="zoom"  hwrid="/zoom"></object>
</object>



"""

import logging

from mxcubecore.BaseHardwareObjects import HardwareObject


class PX2BeamInfo(HardwareObject):
    def __init__(self, *args):
        super().__init__(*args)

        self.beam_position = [328, 220]  # [None, None]
        self.beam_size = [0.010, 0.005]  # [None, None]
        self.shape = "rectangular"

        self.beam_info_dict = {"size_x": None, "size_y": None, "shape": self.shape}

        self.beam_info_dict["size_x"] = 0.010
        self.beam_info_dict["size_y"] = 0.005
        self.beam_info_dict["shape"] = "ellipse"

        # Channels
        self.chanBeamSizeX = None
        self.chanBeamSizeY = None
        self.chanBeamPosX = None
        self.chanBeamPosY = None

        # Zoom motor
        self.zoomMotor = None
        # self.minidiff = None
        self.positionTable = {}

    def init(self):

        try:
            self.chanBeamSizeX = self.get_channel_object("beamsizex")
            self.chanBeamSizeX.connect_signal("update", self.beamSizeXChanged)
        except KeyError:
            logging.getLogger().warning(
                "%s: cannot connect to beamsize x channel ", self.name()
            )

        try:
            self.chanBeamSizeY = self.get_channel_object("beamsizey")
            self.chanBeamSizeY.connect_signal("update", self.beamSizeYChanged)
        except KeyError:
            logging.getLogger().warning(
                "%s: cannot connect to beamsize y channel ", self.name()
            )

        try:
            self.chanBeamPosX = self.get_channel_object("positionx")
            self.chanBeamPosX.connect_signal("update", self.beamPosXChanged)
        except KeyError:
            logging.getLogger().warning(
                "%s: cannot connect to beamposition x channel ", self.name()
            )

        try:
            self.chanBeamPosY = self.get_channel_object("positiony")
            self.chanBeamPosY.connect_signal("update", self.beamPosYChanged)
        except KeyError:
            logging.getLogger().warning(
                "%s: cannot connect to beamposition z channel ", self.name()
            )

        self.zoomMotor = self.get_deviceby_role("zoom")

        self.beam_position[0], self.beam_position[1] = (
            self.chanBeamPosX.value,
            self.chanBeamPosY.value,
        )

        if self.zoomMotor is not None:
            self.connect(
                self.zoomMotor, "predefinedPositionChanged", self.zoomPositionChanged
            )
        else:
            logging.getLogger().info("Zoom - motor is not good ")

    def beamSizeXChanged(self, value):
        logging.getLogger().info("beamSizeX changed. It is %s " % value)
        self.beam_size[0] = value
        self.sizeUpdated()

    def beamSizeYChanged(self, value):
        logging.getLogger().info("beamSizeY changed. It is %s " % value)
        self.beam_size[1] = value
        self.sizeUpdated()

    def beamPosXChanged(self, value):
        logging.getLogger().info("beamPosX changed. It is %s " % value)
        self.beam_position[0] = value
        self.positionUpdated()

    def beamPosYChanged(self, value):
        logging.getLogger().info("beamPosY changed. It is %s " % value)
        self.beam_position[1] = value
        self.positionUpdated()

    def zoomPositionChanged(self, name, offset):
        logging.getLogger().info(
            "zoom position changed. It is %s / offset=%s " % (name, offset)
        )
        self.beam_position[0], self.beam_position[1] = (
            self.chanBeamPosX.value,
            self.chanBeamPosY.value,
        )

    def sizeUpdated(self):
        # TODO check values give by md2 it appears that  beamSizeXChanged beamSize
        self.beam_info_dict["size_x"] = 0.010  # in micro channel in MD2 doesn't work
        self.beam_info_dict["size_y"] = 0.005  #
        self.emit("beamInfoChanged", (self.beam_info_dict,))

    def sizeUpdated2(self):
        # not used
        if None in self.beam_size:
            return
        self.beam_info_dict["size_x"] = self.beam_size[0]
        self.beam_info_dict["size_y"] = self.beam_size[1]

        self.emit("beamInfoChanged", (self.beam_info_dict,))

    def positionUpdated(self):
        self.emit("beamPosChanged", (self.beam_position,))
        self.sizeUpdated()

    def get_beam_info(self):
        # logging.getLogger().warning('returning beam info It is %s ' % str(self.beam_info_dict))
        return self.beam_info_dict

    def get_beam_position(self):
        # logging.getLogger().warning('returning beam positions. It is %s ' % str(self.beam_position))
        return self.beam_position

    def get_beam_size(self):
        """
        Descript. : returns beam size in millimeters
        Return   : list with two integers
        """
        # self.evaluate_beam_info()
        return self.beam_info_dict["size_x"], self.beam_info_dict["size_y"]

    def get_beam_shape(self):
        """
        Descript. :
        Arguments :
        Return    :
        """
        # self.evaluate_beam_info()
        return self.shape

    def get_slit_gaps(self):
        return None, None

    def get_beam_divergence_hor(self):
        return self.get_property("beam_divergence_hor")

    def get_beam_divergence_ver(self):
        return self.get_property("beam_divergence_vert")



import urllib
import xml.etree.cElementTree as et
from io import BytesIO

import requests
from PIL import Image


def get_image(url):
    f = urllib.request.urlopen(url)
    img = f.read()
    return img


def get_image_size(url):
    img_data = requests.get(url).content
    im = Image.open(BytesIO(img_data))
    return (im.size)


class CrimsXtal:
    def __init__(self, *args):
        self.crystal_uuid = ""
        self.pin_id = ""
        self.login = ""
        self.sample = ""
        self.column = 0
        self.id_sample = 0
        self.id_trial = 0
        self.row = ""
        self.shelf = 0
        self.comments = ""
        self.offset_x = 0.0
        self.offset_y = 0.0
        self.shape = ""
        self.image_url = ""
        self.image_rotation = 0.0
        self.summary_url = ""

    def get_address(self):
        return "%s%02d-%d" % (self.row, self.column, self.shelf)

    def get_image(self):
        if len(self.image_url) > 0:
            try:
                if self.image_url.startswith("http://"):
                    self.image_url = "https://" + self.image_url[7]
                image_string = urllib.request.urlopen(self.image_url).read()
                return image_string
            except Exception:
                return

    def get_image_size(self):
        img_data = requests.get(self.image_url).content
        im = Image.open(BytesIO(img_data))
        return (im.size)

    def get_summary_url(self):
        if len(self.summary_url == 0):
            return None
        return self.summary_url


class Plate:
    def __init__(self, *args):
        self.barcode = ""
        self.plate_type = ""
        self.xtal_list = []


class ProcessingPlan:
    def __init__(self, *args):
        self.plate = Plate()


def get_processing_plan(barcode, crims_url, crims_user_agent, harvester_key):
    processing_plan = None
    try:
        xml = None
        url = f"{crims_url}{barcode}/plans/xml"

        headers = {
            'User-Agent' : crims_user_agent,
            'harvester-key' : harvester_key,
        }

        req = urllib.request.Request(url, data=None, headers=headers)

        with urllib.request.urlopen(req) as response:
            xml = response.read()

        tree = et.fromstring(xml)

        processing_plan = ProcessingPlan()
        plate = tree.findall("Plate")[0]

        processing_plan.plate.barcode = plate.find("Barcode").text
        processing_plan.plate.plate_type = plate.find("PlateType").text
        for x in plate.findall("Drop"):
            if (x.find("Pin")):
                xtal = CrimsXtal()
                xtal.pin_id = x.find("Pin").find("PinUUID").text
                xtal.crystal_uuid = x.find("Pin").find("Xtal").find("CrystalUUID").text
                # xtal.label = x.find("Label").text
                # xtal.login = plate.find("Login").text
                xtal.sample = x.find("Sample").text
                xtal.id_sample = int(x.find("idSample").text)
                xtal.column = int(x.find("Column").text)
                xtal.row = x.find("Row").text
                xtal.shelf = int(x.find("Shelf").text)
                # xtal.comments = x.find("Comments").text
                xtal.offset_x = float(x.find("Pin").find("Xtal").find("X").text) / 100.0
                xtal.offset_y = float(x.find("Pin").find("Xtal").find("Y").text) / 100.0
                xtal.shape = x.find("Pin").find("Shape").text
                xtal.image_url = x.find("IMG_URL").text

                xtal.image_height = get_image_size(x.find("IMG_URL").text)[0]
                xtal.image_width = get_image_size(x.find("IMG_URL").text)[1]

                xtal.image_date = x.find("IMG_Date").text
                xtal.image_rotation = float(x.find("ImageRotation").text)
                # xtal.summary_url = x.find("SUMMARY_URL").text
                processing_plan.plate.xtal_list.append(xtal)
        return processing_plan
    except Exception as ex:
        print("Error on getting processing plan because of: %s" % str(ex))
        return processing_plan


def send_data_collection_info_to_crims(
    crims_url: str,
    crystaluuid: str,
    datacollectiongroupid: str,
    dcid: str,
    proposal: str,
    rest_token: str,
    crims_key: str
) -> str:
    try:

        url = (
            f"{crims_url}{crystaluuid}/dcgroupid/{datacollectiongroupid}/dcid/"
            f"{dcid}/mx/{proposal}/token/{rest_token}?janitor_key={crims_key}"
        )

        # data = {
        #     "crystal_uuid": str(crystaluuid),
        #     "datacollectionGroupId": str(datacollectiongroupid),
        #     "dcid": str(dcid),
        #     "mx": str(proposal),
        #     "token": str(rest_token),
        # }
        response = requests.get(url, timeout=900)
        # response = post(url, data=data, timeout=900)
        print(response.text)
        return response.text
    except Exception as ex:
        msg = "POST to %s failed reason %s" % (url, str(ex))
        return msg



from __future__ import print_function

import logging
import sys

from suds import WebFault
from suds.client import Client
from suds.sudsobject import asdict

from mxcubecore.HardwareObjects.abstract.ISPyBAbstractLims import ISPyBAbstractLIMS
from mxcubecore.model.lims_session import LimsSessionManager

try:
    from urllib2 import URLError
    from urlparse import urljoin
except Exception:
    # Python3
    from urllib.error import URLError

suds_encode = str.encode

if sys.version_info > (3, 0):
    suds_encode = bytes.decode

logging.getLogger("suds").setLevel(logging.INFO)


class UserTypeISPyBLims(ISPyBAbstractLIMS):
    """
    ISPyB user-based client
    """

    def __init__(self, name):
        super().__init__(name)
        self._collection = None

    def get_user_name(self):
        return self.user_name

    def is_user_login_type(self):
        return True

    def init(self):
        super().init()
        try:
            # ws_root is a property in the configuration xml file
            if self.ws_root:
                global _WS_COLLECTION_URL

                _WSDL_ROOT = self.ws_root.strip()

                _WS_COLLECTION_URL = _WSDL_ROOT + "ToolsForCollectionWebService?wsdl"

                if self.ws_root.strip().startswith("https://"):
                    from suds.transport.https import HttpAuthenticated
                else:
                    from suds.transport.http import HttpAuthenticated

                t2 = HttpAuthenticated(
                    username=self.ws_username,
                    password=self.ws_password,
                    proxy=self.proxy,
                )

                try:

                    self._collection = Client(
                        _WS_COLLECTION_URL,
                        timeout=3,
                        transport=t2,
                        cache=None,
                        proxy=self.proxy,
                    )

                    # ensure that suds do not create those files in tmp

                    self._collection.set_options(
                        cache=None, location=_WS_COLLECTION_URL
                    )

                except URLError as e:
                    logging.getLogger("ispyb_client").exception(str(e))
                    return
        except Exception as e:
            logging.getLogger("ispyb_client").exception(str(e))
            return

    def login(self, loginID, psd, ldap_connection=None) -> LimsSessionManager:
        login_name = loginID
        proposal_code = ""
        proposal_number = ""
        self.user_name = loginID

        # if translation of the loginID is needed, need to be tested by ESRF
        if self.loginTranslate is True:
            login_name = self._translate(proposal_code, "ldap") + str(proposal_number)

        # Authentication
        if self.authServerType == "ldap":
            logging.getLogger("HWR").debug(
                "Starting LDAP authentication %s" % login_name
            )
            self.login_ok = self.ldap_login(login_name, psd, ldap_connection)
            msg = loginID
            logging.getLogger("HWR").debug("User %s logged in LDAP" % login_name)
        elif self.authServerType == "ispyb":
            logging.getLogger("HWR").debug("ISPyB login")
            self.login_ok, msg = self.ispyb_login(login_name, psd)
        else:
            raise Exception("Authentication server type is not defined")

        if not self.login_ok:
            msg = "%s." % msg.capitalize()
            # refuse Login
            logging.getLogger("HWR").error("ISPyB login not ok")
            raise Exception("Error lims authentication")

        # login succeed, get proposal and sessions
        self.session_manager = self.adapter.get_sessions_by_username(
            loginID, self.beamline_name
        )

        return self.session_manager

    def get_proposals_by_user(self, user_name):
        proposal_list = []
        res_proposal = []

        if self._shipping:
            try:
                proposals = eval(
                    self._shipping.service.findProposalsByLoginName(user_name)
                )
                if proposal_list is not None:
                    for proposal in proposals:
                        if (
                            proposal["type"].upper() in ["MX", "MB"]
                            and proposal not in proposal_list
                        ):
                            proposal_list.append(proposal)
            except WebFault as e:
                proposal_list = []
                logging.getLogger("ispyb_client").exception(e.message)

            proposal_list = newlist = sorted(
                proposal_list, key=lambda k: int(k["proposalId"])
            )

            res_proposal = []
            if len(proposal_list) > 0:
                for proposal in proposal_list:
                    proposal_code = proposal["code"]
                    proposal_number = proposal["number"]

                    # person
                    try:
                        person = self._shipping.service.findPersonByProposal(
                            proposal_code, proposal_number
                        )
                        if not person:
                            person = {}
                    except WebFault as e:
                        logging.getLogger("ispyb_client").exception(str(e))
                        person = {}

                    # lab
                    try:
                        lab = self._shipping.service.findLaboratoryByProposal(
                            proposal_code, proposal_number
                        )
                        if not lab:
                            lab = {}
                    except WebFault as e:
                        logging.getLogger("ispyb_client").exception(str(e))
                        lab = {}

                    # sessions
                    try:
                        res_sessions = (
                            self._collection.service.findSessionsByProposalAndBeamLine(
                                proposal_code, proposal_number, self.beamline_name
                            )
                        )
                        sessions = []
                        for session in res_sessions:
                            if session is not None:
                                try:
                                    session.startDate = datetime.strftime(
                                        session.startDate, "%Y-%m-%d %H:%M:%S"
                                    )
                                    session.endDate = datetime.strftime(
                                        session.endDate, "%Y-%m-%d %H:%M:%S"
                                    )
                                except Exception:
                                    pass
                                sessions.append(utf_encode(asdict(session)))

                    except WebFault as e:
                        logging.getLogger("ispyb_client").exception(str(e))
                        sessions = []

                    res_proposal.append(
                        {
                            "Proposal": proposal,
                            "Person": utf_encode(asdict(person)),
                            "Laboratory": utf_encode(asdict(lab)),
                            "Session": sessions,
                        }
                    )
            else:
                logging.getLogger("ispyb_client").warning(
                    "No proposals for user %s found" % user_name
                )
        else:
            logging.getLogger("ispyb_client").exception(
                "Error in get_proposal: Could not connect to server,"
                + " returning empty proposal"
            )
        return res_proposal

    def get_proposal_by_username(self, username: str):
        proposal_code = ""
        proposal_number = 0

        empty_dict = {
            "Proposal": {},
            "Person": {},
            "Laboratory": {},
            "Session": {},
            "status": {"code": "error"},
        }

        if not self._shipping:
            logging.getLogger("ispyb_client").warning(
                "Error in get_proposal: Could not connect to server,"
                + " returning empty proposal"
            )
            return empty_dict

        try:
            try:
                person = self._shipping.service.findPersonByLogin(
                    username, self.beamline_name
                )
            except WebFault as e:
                logging.getLogger("ispyb_client").warning(str(e))
                person = {}

            try:
                proposal = self._shipping.service.findProposalByLoginAndBeamline(
                    username, self.beamline_name
                )
                if not proposal:
                    logging.getLogger("ispyb_client").warning(
                        "Error in get_proposal: No proposal has been found to  the user, returning empty proposal"
                    )
                    return empty_dict
                proposal_code = proposal.code
                proposal_number = proposal.number
            except WebFault as e:
                logging.getLogger("ispyb_client").warning(str(e))
                proposal = {}

            try:
                lab = self._shipping.service.findLaboratoryByCodeAndNumber(
                    proposal_code, proposal_number
                )
            except WebFault as e:
                logging.getLogger("ispyb_client").warning(str(e))
                lab = {}

            try:
                res_sessions = (
                    self._collection.service.findSessionsByProposalAndBeamLine(
                        proposal_code, proposal_number, self.beamline_name
                    )
                )
                sessions = []

                # Handels a list of sessions
                for session in res_sessions:
                    if session is not None:
                        try:
                            session.startDate = datetime.strftime(
                                session.startDate, "%Y-%m-%d %H:%M:%S"
                            )
                            session.endDate = datetime.strftime(
                                session.endDate, "%Y-%m-%d %H:%M:%S"
                            )
                        except Exception:
                            pass

                        sessions.append(utf_encode(asdict(session)))

            except WebFault as e:
                logging.getLogger("ispyb_client").warning(str(e))
                sessions = []

        except URLError:
            logging.getLogger("ispyb_client").warning(_CONNECTION_ERROR_MSG)
            return empty_dict

        logging.getLogger("ispyb_client").info(str(sessions))
        return {
            "Proposal": utf_encode(asdict(proposal)),
            "Person": utf_encode(asdict(person)),
            "Laboratory": utf_encode(asdict(lab)),
            "Session": sessions,
            "status": {"code": "ok"},
        }
